from ._dcca import DCCA
from ._dcca_barlow_twins import BarlowTwins
from ._dcca_ey import DCCA_EY
from ._dcca_gh import DCCA_GH
from ._dcca_svd import DCCA_SVD
from ._dcca_noi import DCCA_NOI
from ._dcca_sdl import DCCA_SDL
from ._dtcca import DTCCA
from ._dgcca import DGCCA



import itertools
from typing import List, Union

import numpy as np
from scipy.linalg import block_diag
from sklearn.datasets import make_spd_matrix
from sklearn.utils.validation import check_random_state

from cca_zoo.utils import _process_parameter


class LinearSimulatedData:
    # This class generates simulated data for linear classical with multiple views
    def __init__(
        self,
        view_features: List[int],
        latent_dims: int = 1,
        view_sparsity: Union[List[float], float] = None,
        correlation: Union[List[float], float] = 0.99,
        structure: str = "random",
        positive: Union[bool, List[bool]] = False,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        self.view_features = view_features
        self.latent_dims = latent_dims
        self.correlation = correlation
        self.random_state = check_random_state(random_state)
        # Process the correlation parameter to ensure it is a list of length latent_dimensions
        self.correlation = _process_parameter(
            "correlation", correlation, 0.99, latent_dims
        )
        self.view_sparsity = _process_parameter(
            "view_sparsity", view_sparsity, 0.5, len(view_features)
        )
        self.structure = _process_parameter(
            "structure", structure, "random", len(view_features)
        )
        self.positive = _process_parameter(
            "positive", positive, False, len(view_features)
        )
        # Generate the covariance matrices and the true features for each view
        covs, self.true_features = self._generate_covariance_matrices()
        # Generate the joint covariance matrix by combining the view covariances and the cross-covariances
        joint_cov = self._generate_joint_covariance(covs)
        # Compute the Cholesky decomposition of the joint covariance matrix
        self.chol = np.linalg.cholesky(joint_cov)

    def _generate_covariance_matrix(self, view_features, view_structure):
        # Generate a covariance matrix for a single view based on its structure
        if view_structure == "identity":
            # Use an identity matrix as the covariance matrix
            cov = np.eye(view_features)
        elif view_structure == "random":
            # Use a random positive definite matrix as the covariance matrix
            cov = make_spd_matrix(view_features)
        return cov

    def _generate_joint_covariance(self, covs):
        # Generate a joint covariance matrix for all views by stacking the view covariances and adding cross-covariances
        cov = block_diag(*covs)
        splits = np.concatenate(([0], np.cumsum(self.view_features)))
        for i, j in itertools.combinations(range(len(splits) - 1), 2):
            cross = np.zeros((self.view_features[i], self.view_features[j]))
            for _ in range(self.latent_dims):
                # Compute the cross-covariance matrix for a pair of views and a latent dimension using the correlation coefficient and the true features
                A = self.correlation[_] * np.outer(
                    self.true_features[i][:, _], self.true_features[j][:, _]
                )
                # Multiply the cross-covariance matrix by the view covariances to get the final cross-covariance matrix
                cross += covs[i] @ A @ covs[j]
            # Assign the cross-covariance matrix to the corresponding blocks in the joint covariance matrix
            cov[
                splits[i] : splits[i] + self.view_features[i],
                splits[j] : splits[j] + self.view_features[j],
            ] = cross
            cov[
                splits[j] : splits[j] + self.view_features[j],
                splits[i] : splits[i] + self.view_features[i],
            ] = cross.T
        return cov

    def _generate_covariance_matrices(self):
        # Generate a list of covariance matrices and true features for each view using list comprehensions
        covs = [
            self._generate_covariance_matrix(view_features, structure)
            for view_features, structure in zip(self.view_features, self.structure)
        ]

        true_features = [
            self._generate_true_feature(cov, sparsity, view_positive)
            for cov, sparsity, view_positive, view_features in zip(
                covs, self.view_sparsity, self.positive, self.view_features
            )
        ]
        return covs, true_features

    def _generate_true_feature(self, cov, sparsity, view_positive):
        # Generate a true feature matrix for a single view using its covariance matrix and sparsity level

        # Generate a random weight matrix with normal distribution and the same shape as the covariance matrix
        weights = self._generate_weights_matrix(cov.shape[0])

        # Apply a sparsity mask to the weight matrix to make some elements zero
        weights = self._apply_sparsity_mask(weights, sparsity)

        # Make the weight matrix positive if needed by taking the absolute value of the elements
        if view_positive:
            weights = self._make_weights_positive(weights)

        # Decorrelate the weight matrix from the covariance matrix and normalize it
        weights = _decorrelate_dims(weights, cov)
        weights /= np.sqrt(np.diag((weights.T @ cov @ weights)))
        return weights

    def _generate_weights_matrix(self, view_features):
        # Generate a random matrix with normal distribution and the given number of rows and latent dimensions
        return self.random_state.randn(view_features, self.latent_dims)

    def _apply_sparsity_mask(self, weights, sparsity):
        view_features = weights.shape[0]
        # convert sparsity to an integer number of nonzero elements
        if sparsity <= 1:
            sparsity = np.ceil(sparsity * view_features).astype("int")
        # create a binary mask with sparsity number of ones
        mask = np.stack(
            (
                np.concatenate(
                    ([0] * (view_features - sparsity), [1] * sparsity)
                ).astype(bool),
            )
            * self.latent_dims,
            axis=0,
        ).T
        # shuffle the mask randomly
        mask = mask.flatten()
        self.random_state.shuffle(mask)
        mask = mask.reshape(weights.shape)
        # apply the mask to the weights matrix
        return weights * mask

    def _make_weights_positive(self, weights):
        # set all negative elements to zero
        return np.abs(weights)

    def sample(self, n: int):
        X = np.zeros((n, self.chol.shape[0]))
        for i in range(n):
            X[i, :] = self._chol_sample(self.chol)
        views = np.split(X, np.cumsum(self.view_features)[:-1], axis=1)
        return views

    def _chol_sample(self, chol):
        rng = check_random_state(self.random_state)
        return chol @ rng.randn(chol.shape[0])


def _decorrelate_dims(up, cov):
    A = up.T @ cov @ up
    for k in range(1, A.shape[0]):
        up[:, k:] -= np.outer(up[:, k - 1], A[k - 1, k:] / A[k - 1, k - 1])
        A = up.T @ cov @ up
    return up



from ._mcca import MCCA, CCA, rCCA
from ._pls import PLS, MPLS
from ._gcca import GCCA
from ._grcca import GRCCA
from ._ncca import NCCA
from ._partialcca import PartialCCA
from ._prcca import PRCCA
from ._tcca import TCCA
from ._pcacca import PCACCA
from ._kcca import KCCA, KGCCA, KTCCA


__all__ = [
    "CCA",
    "GCCA",
    "GRCCA",
    "KCCA",
    "MCCA",
    "NCCA",
    "PLS",
    "PRCCA",
    "KTCCA",
    "TCCA",
    "PartialCCA",
    "rCCA",
    "PCACCA",
    "KGCCA",
    "MPLS",
]

# if pytorch-lightning is installed then import ._iterative

try:
    from ._iterative import (
        CCAEY,
        CCAGH,
        CCASVD,
        PLS_ALS,  # SCCA_ADMM,; SWCCA,
        PLSEY,
        PLSSVD,
        SCCA_IPLS,
        SCCA_PMD,
        AltMaxVar,
        ElasticCCA,
        PLSStochasticPower,
        SCCA_Parkhomenko,
        SCCA_Span,
    )

    __all__ += [
        "ElasticCCA",
        "SCCA_IPLS",
        "PLS_ALS",
        "SCCA_PMD",
        "SCCA_Parkhomenko",
        "SCCA_Span",
        "AltMaxVar",
        "CCAEY",
        "PLSEY",
        "CCAGH",
        "CCASVD",
        "PLSSVD",
    ]
except:
    # let user know that if they want to use iterative methods they need to install pytorch-lightning from version 2.0.0
    print(
        "To use iterative methods please install pytorch-lightning from version 2.0.0"
    )

classes = __all__



import torch

from cca_zoo.classical._iterative._ey import CCAEY, EYLoop


class CCAGH(CCAEY):
    def _get_module(self, weights=None, k=None):
        return GHALoop(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
            objective="cca",
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class GHALoop(EYLoop):
    def get_AB(self, z):
        latent_dims = z[0].shape[1]
        A = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += torch.cov(zi.T)  # add the auto-covariance of each view to B
                A += torch.cov(torch.hstack((zi, zj)).T)[
                    latent_dims:, :latent_dims
                ]  # add the cross-covariance of each pair of views to A
        return A / len(z), B / len(
            z
        )  # return the normalized matrices (divided by the number of views)

    def loss(self, views, views2=None, **kwargs):
        # Encoding the views with the forward method
        z = self(views)
        # Getting A and B matrices from z
        A, B = self.get_AB(z)
        if views2 is None:
            # Computing rewards and penalties using A and B only
            rewards = torch.trace(2 * A)
            penalties = torch.trace(A @ B)
        else:
            # Encoding another set of views with the forward method
            z2 = self(views2)
            # Getting A' and B' matrices from z2
            _, B_ = self.get_AB(z2)
            # Computing rewards and penalties using A and B'
            rewards = torch.trace(2 * A)
            penalties = torch.trace(A @ B_)
        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }



import warnings
from typing import Iterable, Union

import numpy as np
import torch
from sklearn.linear_model import ElasticNet, Lasso, Ridge, SGDRegressor

from cca_zoo.classical._iterative._base import BaseLoop
from cca_zoo.classical._iterative._deflation import BaseDeflation
from cca_zoo.utils import _process_parameter


class ElasticCCA(BaseDeflation):
    r"""
    A class used to fit an elastic CCA model for two or more views of data.

    This model finds the linear projections of multiple views that maximize their pairwise correlations while enforcing L1 and L2 penalties on the projection vectors.

    ElasticCCA uses CCA with an auxiliary variable target i.e. MAXVAR configuration

    .. math::

        w_{opt}, t_{opt}=\underset{w,t}{\mathrm{argmin}}\{\sum_i \|X_iw_i-t\|^2 + \alpha_i\|w_i\|^2_2 + \text{l1_ratio}\|w_i\|_1\}\\

        \text{subject to:}

        t^Tt=n

    Where :math:`\alpha_i` is the constant that multiplies the penalty terms for view i and :math:`\text{l1_ratio}` is the ratio between L1 and L2 penalties.

    Parameters
    ----------
    latent_dimensions : int, default=1
        Number of latent dimensions to use
    copy_data : bool, default=True
        Whether to copy the data or overwrite it
    random_state : int, default=None
        Random seed for initialization
    deflation : str, default="cca"
        Whether to use CCA or PLS deflation
    initialization : str or callable, default="pls"
        How to initialize the weights. Can be "pls" or "random" or a callable
    tol : float, default=1e-3
        Tolerance for convergence
    alpha : float or list of floats, default=None
        Regularisation parameter for the L2 penalty. If None, defaults to 1.0
    l1_ratio : float or list of floats, default=None
        Regularisation parameter for the L1 penalty. If None, defaults to 0.0
    stochastic : bool, default=False
        Whether to use stochastic gradient descent
    positive : bool or list of bools, default=None
        Whether to use non-negative constraints

    Examples
    --------
    >>> from cca_zoo.classical import ElasticCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = ElasticCCA(c=[1e-1,1e-1],l1_ratio=[0.5,0.5], random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.9316638])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        epochs: int = 100,
        deflation="cca",
        initialization: Union[str, callable] = "pls",
        tol: float = 1e-3,
        alpha: Union[Iterable[float], float] = None,
        l1_ratio: Union[Iterable[float], float] = None,
        stochastic=False,
        positive: Union[Iterable[bool], bool] = None,
        convergence_checking=None,
        track=None,
        verbose=False,
    ):
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.stochastic = stochastic
        self.positive = positive
        if self.positive is not None and stochastic:
            self.stochastic = False
            warnings.warn(
                "Non negative constraints cannot be used with _stochastic regressors. Switching to _stochastic=False"
            )
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            epochs=epochs,
            deflation=deflation,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            convergence_checking=convergence_checking,
            patience=0,
            track=track,
            verbose=verbose,
        )

    def _check_params(self):
        self.alpha = _process_parameter("alpha", self.alpha, 0, self.n_views_)
        self.l1_ratio = _process_parameter("l1_ratio", self.l1_ratio, 0, self.n_views_)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )

    def _get_module(self, weights=None, k=None):
        return ElasticLoop(
            weights=weights,
            k=k,
            alpha=self.alpha,
            l1_ratio=self.l1_ratio,
            stochastic=self.stochastic,
            positive=self.positive,
            tol=self.tol,
            random_state=self.random_state,
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True}


class ElasticLoop(BaseLoop):
    def __init__(
        self,
        weights,
        k=None,
        alpha=None,
        l1_ratio=None,
        stochastic=False,
        positive=None,
        tol=1e-3,
        random_state=None,
        tracking=False,
        convergence_checking=False,
    ):
        super().__init__(
            weights=weights,
            k=k,
            automatic_optimization=False,
            tracking=tracking,
            convergence_checking=convergence_checking,
        )
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.n_views = len(self.weights)
        self.n_samples_ = self.weights[0].shape[0]
        self.regressors = initialize_regressors(
            alpha,
            l1_ratio,
            positive,
            stochastic,
            tol,
            random_state,
        )
        for view_index, weight in enumerate(weights):
            self.regressors[view_index].coef_ = weight[:, 0]
            self.regressors[view_index].intercept_ = 0

    def forward(self, views: list) -> list:
        scores = []
        for view_index, view in enumerate(views):
            scores.append(self.regressors[view_index].predict(view))
        return scores

    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        old_weights = self.weights.copy()
        # Update each view using loop update function
        for view_index, view in enumerate(batch["views"]):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            target /= np.linalg.norm(target) / np.sqrt(self.n_samples_)
            # Solve the elastic regression
            self.regressors[view_index] = self.regressors[view_index].fit(
                batch["views"][view_index], target
            )
            self.weights[view_index] = self.regressors[view_index].coef_
        # if tracking or convergence_checking is enabled, compute the objective function
        if self.tracking or self.convergence_checking:
            objective = self.objective(batch["views"])
            # check that the maximum change in weights is smaller than the tolerance times the maximum absolute value of the weights
            weights_change = torch.tensor(
                np.max(
                    [
                        np.max(np.abs(old_weights[i] - self.weights[i]))
                        / np.max(np.abs(self.weights[i]))
                        for i in range(len(self.weights))
                    ]
                )
            )
            return {"loss": torch.tensor(objective), "weights_change": weights_change}

    def on_fit_end(self) -> None:
        self.weights = [regressor.coef_ for regressor in self.regressors]

    def objective(self, views):
        scores = np.stack(self(views))
        objective = 0
        for view_index, view in enumerate(views):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            target /= np.linalg.norm(target) / np.sqrt(self.n_samples_)
            objective += elastic_objective(
                view,
                self.regressors[view_index].coef_,
                target,
                self.alpha[view_index],
                self.l1_ratio[view_index],
            )
        return objective


class SCCA_IPLS(ElasticCCA):
    r"""
    A class used to fit a sparse CCA model by iterative rescaled lasso regression. Implemented by ElasticCCA with l1 ratio=1 and SUMCOR form.

    This model finds the linear projections of multiple views that maximize their pairwise correlations while enforcing sparsity constraints on the projection vectors.

    The objective function of sparse CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmin}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \sum_i \alpha_i\|w_i\|_1 \}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=n

    where :math:`\alpha_i` are the sparsity parameters for each view.

    This model uses the SUMCOR form of CCA, which means that the projection vectors are normalized by their auto-covariance matrices.



    """

    def _get_module(self, weights=None, k=None):
        self.l1_ratio = [1] * self.n_views_
        return IPLSLoop(
            weights=weights,
            k=k,
            alpha=self.alpha,
            l1_ratio=self.l1_ratio,
            stochastic=self.stochastic,
            positive=self.positive,
            tol=self.tol,
            random_state=self.random_state,
        )


class IPLSLoop(ElasticLoop):
    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        old_weights = self.weights.copy()
        # Update each view using loop update function
        for view_index, view in enumerate(batch["views"]):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            # Solve the elastic regression
            self.regressors[view_index] = self.regressors[view_index].fit(
                batch["views"][view_index], target
            )
            self.regressors[view_index].coef_ /= np.linalg.norm(
                view @ self.regressors[view_index].coef_
            ) / np.sqrt(self.n_samples_)
            self.weights[view_index] = self.regressors[view_index].coef_
            # if tracking or convergence_checking is enabled, compute the objective function
            if self.tracking or self.convergence_checking:
                objective = self.objective(batch["views"])
                # check that the maximum change in weights is smaller than the tolerance times the maximum absolute value of the weights
                weights_change = torch.tensor(
                    np.max(
                        [
                            np.max(np.abs(old_weights[i] - self.weights[i]))
                            / np.max(np.abs(self.weights[i]))
                            for i in range(len(self.weights))
                        ]
                    )
                )
                return {"loss": objective, "weights_change": weights_change}

    def objective(self, views):
        scores = np.stack(self(views))

        objective = 0
        for view_index, view in enumerate(views):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            target /= np.linalg.norm(target) / np.sqrt(self.n_samples_)
            objective += elastic_objective(
                view,
                self.regressors[view_index].coef_,
                target,
                self.alpha[view_index],
                self.l1_ratio[view_index],
            )
        return {"loss": torch.tensor(objective)}


def elastic_objective(x, w, y, alpha, l1_ratio):
    n = len(y)
    z = x @ w
    objective = np.linalg.norm(z - y) ** 2 / (2 * n)
    l1_pen = alpha * l1_ratio * np.linalg.norm(w, ord=1)
    l2_pen = alpha * (1 - l1_ratio) * np.linalg.norm(w, ord=2)
    return objective + l1_pen + l2_pen


def initialize_regressors(alpha, l1_ratio, positive, stochastic, tol, random_state):
    regressors = []
    for alpha, l1_ratio, positive in zip(alpha, l1_ratio, positive):
        if stochastic:
            regressors.append(
                SGDRegressor(
                    penalty="elasticnet",
                    alpha=alpha,
                    l1_ratio=l1_ratio,
                    fit_intercept=False,
                    tol=tol,
                    warm_start=True,
                    random_state=random_state,
                )
            )
        elif l1_ratio == 0:
            regressors.append(
                Ridge(
                    alpha=alpha,
                    fit_intercept=False,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                )
            )
        elif l1_ratio == 1:
            regressors.append(
                Lasso(
                    alpha=alpha,
                    fit_intercept=False,
                    warm_start=True,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                    selection="random",
                )
            )
        else:
            regressors.append(
                ElasticNet(
                    alpha=alpha,
                    l1_ratio=l1_ratio,
                    fit_intercept=False,
                    warm_start=True,
                    positive=positive,
                    random_state=random_state,
                    tol=tol,
                    selection="random",
                )
            )
    return regressors



from typing import Union

import numpy as np
import torch

from cca_zoo.classical._iterative._base import BaseGradientLoop, BaseIterative
from cca_zoo.classical._pls import PLSMixin


class CCAEY(BaseIterative):
    """
    A class used to fit Regularized CCA by Delta-EigenGame

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A GeneralizedDeflation EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=None,
        epochs=100,
        learning_rate=1e-1,
        initialization: Union[str, callable] = "random",
        dataloader_kwargs=None,
        optimizer_kwargs=None,
        convergence_checking=None,
        patience=10,
        track=None,
        verbose=False,
    ):
        self.optimizer_kwargs = optimizer_kwargs
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            tol=tol,
            batch_size=batch_size,
            epochs=epochs,
            learning_rate=learning_rate,
            initialization=initialization,
            dataloader_kwargs=dataloader_kwargs,
            convergence_checking=convergence_checking,
            patience=patience,
            track=track,
            verbose=verbose,
        )
        # ensure dataloader_kwargs['shuffle'] is True
        self.dataloader_kwargs["shuffle"] = True

    def _get_module(self, weights=None, k=None):
        return EYLoop(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
            objective="cca",
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class PLSEY(CCAEY, PLSMixin):
    """
    A class used to fit Regularized CCA by Delta-EigenGame

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A GeneralizedDeflation EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=None,
        epochs=1,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        dataloader_kwargs=None,
        convergence_checking=False,
        patience=10,
        track=False,
        verbose=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            tol=tol,
            batch_size=batch_size,
            epochs=epochs,
            learning_rate=learning_rate,
            initialization=initialization,
            dataloader_kwargs=dataloader_kwargs,
            convergence_checking=convergence_checking,
            patience=patience,
            track=track,
            verbose=verbose,
        )
        self.previous_views = None

    def _get_module(self, weights=None, k=None):
        return EYLoop(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
            objective="pls",
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class EYLoop(BaseGradientLoop):
    def __init__(
        self,
        weights=None,
        k=None,
        learning_rate=1e-3,
        optimizer_kwargs=None,
        objective="cca",
        tracking=False,
        convergence_checking=False,
    ):
        super().__init__(
            weights=weights,
            k=k,
            learning_rate=learning_rate,
            optimizer_kwargs=optimizer_kwargs,
            tracking=tracking,
            convergence_checking=convergence_checking,
        )
        self.objective = objective
        self.batch_queue = []
        self.val_batch_queue = []

    def training_step(self, batch, batch_idx):
        # Checking if the queue has at least one batch
        if len(self.batch_queue) < 1:
            # Adding the current batch to the queue
            self.batch_queue.append(batch)
            # Returning a zero loss
            loss = {
                "loss": torch.tensor(0, requires_grad=True, dtype=torch.float32),
            }
        else:
            # randomly select a batch from the queue
            batch2 = self.batch_queue[np.random.randint(0, len(self.batch_queue))]
            # Computing the loss with the current batch and the oldest batch in the queue
            loss = self.loss(batch["views"], batch2["views"])
            # Adding the current batch to the queue and removing the oldest batch
            self.batch_queue.append(batch)
            self.batch_queue.pop(0)
        # Logging the loss components
        for k, v in loss.items():
            self.log(k, v, prog_bar=False)
        return loss

    def get_AB(self, z):
        latent_dims = z[0].shape[1]
        A = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            latent_dims, latent_dims, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    if self.objective == "cca":
                        B += torch.cov(
                            zi.T
                        )  # add the auto-covariance of each view to B
                    elif self.objective == "pls":
                        B += self.weights[i].T @ self.weights[i]
                else:
                    A += torch.cov(torch.hstack((zi, zj)).T)[
                        latent_dims:, :latent_dims
                    ]  # add the cross-covariance of each pair of views to A
        return A / len(z), B / len(
            z
        )  # return the normalized matrices (divided by the number of views)

    def loss(self, views, views2=None, **kwargs):
        # Encoding the views with the forward method
        z = self(views)
        # Getting A and B matrices from z
        A, B = self.get_AB(z)

        if views2 is None:
            # Computing rewards and penalties using A and B only
            rewards = torch.trace(2 * A)
            penalties = torch.trace(B @ B)

        else:
            # Encoding another set of views with the forward method
            z2 = self(views2)
            # Getting A' and B' matrices from z2
            A_, B_ = self.get_AB(z2)
            # Computing rewards and penalties using A and B'
            rewards = torch.trace(2 * A)
            penalties = torch.trace(B @ B_)

        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }



import itertools
import warnings

import numpy as np
import torch

from cca_zoo.classical._iterative._base import BaseLoop
from cca_zoo.classical._iterative._deflation import BaseDeflation
from cca_zoo.classical._pls import PLSMixin
from cca_zoo.classical._search import _delta_search
from cca_zoo.utils import _check_converged_weights, _process_parameter


class SCCA_PMD(BaseDeflation, PLSMixin):
    r"""
    A class used to fit a sparse CCA model by penalized matrix decomposition (PMD).

    This model finds the linear projections of two views that maximize their correlation while enforcing sparsity constraints on the projection vectors.

    The objective function of SCCA-PMD is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        \|w_i\|_2^2=1

        \|w_i\|_1\leq \tau_i

    where :math:`\tau_i` are the sparsity parameters for each view.

    The algorithm alternates between updating the weights for each view and applying a soft thresholding operator to enforce the sparsity constraints.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random seed for reproducibility, by default None
    epochs : int, optional
        Number of iterations to run the algorithm, by default 100
    deflation : str, optional
        Deflation scheme to use, by default "cca"
    initialization : str, optional
        Initialization scheme to use, by default "pls"
    tol : float, optional
        Tolerance for convergence, by default 1e-3
    positive : Union[Iterable[bool], bool], optional
        Whether to enforce positivity constraints on the weights, by default False
    tau : Union[Iterable[float], float], optional
        Sparsity parameter or list of parameters for each view, by default None. If None, it will be set to 1 for each view.
    convergence_checking : str, optional
        Convergence scheme to use, by default None
    track : Union[Iterable[str], str], optional
        List of metrics to track during training, by default None
    verbose : bool, optional
        Whether to print progress, by default False
    """

    def __init__(
        self,
        latent_dimensions=1,
        copy_data=True,
        random_state=None,
        epochs=100,
        deflation="cca",
        initialization="pls",
        tol=1e-3,
        positive=False,
        tau=None,
        convergence_checking=None,
        track=None,
        verbose=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            epochs=epochs,
            deflation=deflation,
            initialization=initialization,
            tol=tol,
            convergence_checking=convergence_checking,
            patience=0,
            track=track,
            verbose=verbose,
        )
        self.tau = tau
        self.positive = positive

    def _check_params(self):
        if self.tau is None:
            warnings.warn(
                "tau parameter not set. Setting to tau=1 i.e. maximum regularisation of l1 norm"
            )
        self.tau = _process_parameter("tau", self.tau, 1, self.n_views_)
        if any(tau < 0 or tau > 1 for tau in self.tau):
            raise ValueError(
                "All regularisation parameters should be between 0 and 1 "
                f"1. tau=[{self.tau}]"
            )
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )

    def _get_module(self, weights=None, k=None):
        return PMDLoop(
            weights=weights,
            k=k,
            tau=self.tau,
            tol=self.tol,
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True}


class PMDLoop(BaseLoop):
    def __init__(
        self,
        weights,
        k=None,
        tau=None,
        tol=1e-3,
        tracking=False,
        convergence_checking=False,
    ):
        super().__init__(
            weights=weights,
            k=k,
            tracking=tracking,
            convergence_checking=convergence_checking,
        )
        self.tau = tau
        self.tol = tol
        shape_sqrts = [np.sqrt(weight.shape[0]) for weight in self.weights]
        self.t = [max(1, x * y) for x, y in zip(self.tau, shape_sqrts)]

    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        old_weights = self.weights.copy()
        # Update each view using loop update function
        for view_index, view in enumerate(batch["views"]):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            self.weights[view_index] = batch["views"][view_index].T @ target
            self.weights[view_index] = _delta_search(
                self.weights[view_index],
                self.t[view_index],
                tol=self.tol,
            )
            _check_converged_weights(self.weights[view_index], view_index)
        # if tracking or convergence_checking is enabled, compute the objective function
        if self.tracking or self.convergence_checking:
            objective = self.objective(batch["views"])
            # check that the maximum change in weights is smaller than the tolerance times the maximum absolute value of the weights
            weights_change = torch.tensor(
                np.max(
                    [
                        np.max(np.abs(old_weights[i] - self.weights[i]))
                        / np.max(np.abs(self.weights[i]))
                        for i in range(len(self.weights))
                    ]
                )
            )
            return {"loss": torch.tensor(objective), "weights_change": weights_change}

    def objective(self, views):
        transformed_views = self(views)
        all_covs = []
        # sum all the pairwise covariances except self covariance
        for i, j in itertools.combinations(range(len(transformed_views)), 2):
            if i != j:
                all_covs.append(
                    np.cov(
                        np.hstack(
                            (
                                transformed_views[i],
                                transformed_views[j],
                            )
                        ).T
                    )
                )
        return np.sum(all_covs)



import numpy as np
import scipy.sparse as sp
from sklearn.utils.fixes import loguniform
from sklearn.utils.validation import check_random_state

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.model_selection import GridSearchCV, RandomizedSearchCV
from cca_zoo.classical import (
    CCA,
    GCCA,
    GRCCA,
    KCCA,
    MCCA,
    NCCA,  # SCCA_ADMM,
    PLS,
    PRCCA,
    SCCA_IPLS,
    SCCA_PMD,
    AltMaxVar,
    ElasticCCA,
    PartialCCA,
    SCCA_Parkhomenko,
    SCCA_Span,
    rCCA,
    MPLS,
)

n = 50
rng = check_random_state(0)
X = rng.rand(n, 10)
Y = rng.rand(n, 11)
Z = rng.rand(n, 12)
X_sp = sp.random(n, 10, density=0.5, random_state=rng)
Y_sp = sp.random(n, 11, density=0.5, random_state=rng)
# centre the data
X -= X.mean(axis=0)
Y -= Y.mean(axis=0)
Z -= Z.mean(axis=0)
X_sp -= X_sp.mean(axis=0)
Y_sp -= Y_sp.mean(axis=0)


def test_linear_simulated_data():
    sim_data = LinearSimulatedData([10, 10]).sample(100)
    assert CCA().fit(sim_data).score(sim_data) > 0.9


def test_regularized_methods():
    # Test that linear regularized methods match PLS solution when using maximum regularisation.
    latent_dims = 2
    c = 1
    rcca = rCCA(latent_dimensions=latent_dims, c=[c, c]).fit([X, Y])
    mcca = MCCA(latent_dimensions=latent_dims, c=[c, c]).fit([X, Y])
    pls = PLS(latent_dimensions=latent_dims).fit([X, Y])
    gcca = GCCA(latent_dimensions=latent_dims, c=[c, c]).fit([X, Y])
    kernel = KCCA(
        latent_dimensions=latent_dims, c=[c, c], kernel=["linear", "linear"]
    ).fit((X, Y))
    mpls = MPLS(latent_dimensions=latent_dims).fit([X, Y])
    corr_gcca = gcca.score((X, Y))
    corr_mcca = mcca.score((X, Y))
    corr_kernel = kernel.score((X, Y))
    corr_pls = pls.score((X, Y))
    corr_rcca = rcca.score((X, Y))
    corr_mpls = mpls.score((X, Y))
    # Check the correlations from each unregularized method are the same
    assert np.testing.assert_array_almost_equal(corr_pls, corr_mcca, decimal=1) is None
    assert (
        np.testing.assert_array_almost_equal(corr_pls, corr_kernel, decimal=1) is None
    )
    assert np.testing.assert_array_almost_equal(corr_pls, corr_rcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_pls, corr_mpls, decimal=1) is None


def test_sparse_methods():
    tau1 = [5e-1]
    tau2 = [5e-1]
    param_grid = {"tau": [tau1, tau2]}
    pdd_cv = GridSearchCV(
        AltMaxVar(proximal="L0", random_state=rng), param_grid=param_grid
    ).fit([X, Y])
    alpha1 = loguniform(1e-2, 2e-2)
    alpha2 = loguniform(1e-2, 2e-2)
    param_grid = {"alpha": [alpha1, alpha2], "l1_ratio": [[0.9], [0.9]]}
    elastic_cv = RandomizedSearchCV(
        ElasticCCA(random_state=rng), param_distributions=param_grid, n_iter=4
    ).fit([X, Y])
    alpha1 = loguniform(1e-2, 2e-2)
    alpha2 = loguniform(1e-2, 2e-2)
    param_grid = {"alpha": [alpha1, alpha2]}
    scca_cv = RandomizedSearchCV(
        SCCA_IPLS(random_state=rng), param_distributions=param_grid
    ).fit([X, Y])
    tau1 = [0.1]
    tau2 = [0.1]
    param_grid = {"tau": [tau1, tau2]}
    pmd_cv = GridSearchCV(SCCA_PMD(random_state=rng), param_grid=param_grid).fit([X, Y])
    tau1 = [2e-1]
    tau2 = [2e-1]
    param_grid = {"tau": [tau1, tau2]}
    parkhomenko_cv = GridSearchCV(
        SCCA_Parkhomenko(random_state=rng), param_grid=param_grid
    ).fit([X, Y])
    tau1 = [2e-2]
    tau2 = [1e-2]
    param_grid = {"tau": [tau1, tau2]}
    # admm_cv = GridSearchCV(SCCA_ADMM(random_state=rng), param_grid=param_grid).fit(
    #     [X, Y]
    # )
    assert (pdd_cv.best_estimator_.weights[0] == 0).sum() > 0
    assert (pdd_cv.best_estimator_.weights[1] == 0).sum() > 0
    assert (pmd_cv.best_estimator_.weights[0] == 0).sum() > 0
    assert (pmd_cv.best_estimator_.weights[1] == 0).sum() > 0
    assert (scca_cv.best_estimator_.weights[0] == 0).sum() > 0
    assert (scca_cv.best_estimator_.weights[1] == 0).sum() > 0
    # assert (admm_cv.best_estimator_.weights[0] == 0).sum() > 0
    # assert (admm_cv.best_estimator_.weights[1] == 0).sum() > 0
    assert (parkhomenko_cv.best_estimator_.weights[0] == 0).sum() > 0
    assert (parkhomenko_cv.best_estimator_.weights[1] == 0).sum() > 0
    assert (elastic_cv.best_estimator_.weights[0] == 0).sum() > 0
    assert (elastic_cv.best_estimator_.weights[1] == 0).sum() > 0


def test_weighted_GCCA_methods():
    # TODO we have view weighted GCCA and missing observation GCCA
    latent_dims = 2
    c = 0
    unweighted_gcca = GCCA(latent_dimensions=latent_dims, c=[c, c]).fit([X, Y])
    deweighted_gcca = GCCA(
        latent_dimensions=latent_dims, c=[c, c], view_weights=[0.5, 0.5]
    ).fit([X, Y])
    corr_unweighted_gcca = unweighted_gcca.score((X, Y))
    corr_deweighted_gcca = deweighted_gcca.score((X, Y))
    # Check the correlations from each unregularized method are the same
    K = np.ones((2, X.shape[0]))
    K[0, 200:] = 0
    unobserved_gcca = GCCA(latent_dimensions=latent_dims, c=[c, c]).fit((X, Y), K=K)
    assert (
        np.testing.assert_array_almost_equal(
            corr_unweighted_gcca, corr_deweighted_gcca, decimal=1
        )
        is None
    )


def test_NCCA():
    latent_dims = 1
    ncca = NCCA(latent_dimensions=latent_dims).fit((X, Y))
    corr_ncca = ncca.score((X, Y))
    assert corr_ncca > 0.9


def test_l0():
    span_cca = SCCA_Span(
        latent_dimensions=1, regularisation="l0", tau=[2, 2], random_state=rng
    ).fit([X, Y])
    # swcca = SWCCA(tau=[5, 5], sample_support=5, random_state=rng).fit([X, Y])
    assert (np.abs(span_cca.weights[0]) > 1e-5).sum() == 2
    assert (np.abs(span_cca.weights[1]) > 1e-5).sum() == 2
    # assert (np.abs(swcca.weights[0]) > 1e-5).sum() == 5
    # assert (np.abs(swcca.weights[1]) > 1e-5).sum() == 5
    # assert (np.abs(swcca.sample_weights) > 1e-5).sum() == 5


def test_partialcca():
    # Tests that partial CCA scores are not correlated with partials
    pcca = PartialCCA(latent_dimensions=3)
    pcca.fit((X, Y), partials=Z)
    assert np.allclose(
        np.corrcoef(pcca.transform((X, Y), partials=Z)[0], Z, rowvar=False)[:3, :3]
        - np.eye(3),
        0,
        atol=0.001,
    )


def test_PRCCA():
    # Test that PRCCA works
    prcca = PRCCA(latent_dimensions=2, c=[0, 0]).fit(
        (X, Y), idxs=(np.arange(10), np.arange(11))
    )
    cca = CCA(latent_dimensions=2).fit([X, Y])
    assert (
        np.testing.assert_array_almost_equal(
            cca.score((X, Y)), prcca.score((X, Y)), decimal=1
        )
        is None
    )
    prcca = PRCCA(latent_dimensions=2, c=[1, 1]).fit(
        (X, Y), idxs=(np.arange(10), np.arange(11))
    )
    pls = PLS(latent_dimensions=2).fit([X, Y])
    assert (
        np.testing.assert_array_almost_equal(
            pls.score((X, Y)), prcca.score((X, Y)), decimal=1
        )
        is None
    )


def test_GRCCA():
    feature_group_1 = np.zeros(X.shape[1], dtype=int)
    feature_group_1[:3] = 1
    feature_group_1[3:6] = 2
    feature_group_2 = np.zeros(Y.shape[1], dtype=int)
    feature_group_2[:3] = 1
    feature_group_2[3:6] = 2
    feature_group_3 = np.zeros(Z.shape[1], dtype=int)
    feature_group_3[:3] = 1
    feature_group_3[3:6] = 2
    # Test that GRCCA works
    grcca = GRCCA(latent_dimensions=1, c=[100, 0], mu=0).fit(
        (X, Y), feature_groups=[feature_group_1, feature_group_2]
    )
    grcca.score((X, Y))
    grcca.transform((X, Y))
    grcca = GRCCA(c=[100, 0, 50]).fit(
        (X, Y, Z), feature_groups=[feature_group_1, feature_group_2, feature_group_3]
    )


def test_deflation():
    # test deflation works with pls and cca using SCCA_PMD
    ccamodel = SCCA_PMD(latent_dimensions=2, tau=1, random_state=rng, deflation="cca")
    ccamodel.fit([X, Y])
    pls = PLS(latent_dimensions=2)
    pls.fit([X, Y])
    plsmodel = SCCA_PMD(latent_dimensions=2, tau=1, random_state=rng, deflation="pls")
    plsmodel.fit([X, Y])
    assert (
        np.testing.assert_array_almost_equal(
            pls.score((X, Y)), plsmodel.score((X, Y)), decimal=1
        )
        is None
    )



__all__ = [
    "data",
    "model_selection",
    "classical",
    "visualisation",
]

# if can import deep add to all
try:
    import cca_zoo.deep

    __all__.append("deep")
except ModuleNotFoundError:
    pass
try:
    import cca_zoo.probabilistic

    __all__.append("probabilistic")
except ModuleNotFoundError:
    pass



from pytorch_lightning import Callback, LightningModule, Trainer


class CorrelationCallback(Callback):
    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        pl_module.log(
            "val/corr",
            pl_module.correlation(trainer.val_dataloaders[0]).sum(),
        )



from typing import Union

import numpy as np


from cca_zoo.classical._iterative._base import BaseLoop
from cca_zoo.classical._iterative._deflation import BaseDeflation
from cca_zoo.classical._pls import PLSMixin


class PLS_ALS(BaseDeflation, PLSMixin):
    r"""
    A class used to fit a PLS model by alternating least squares (ALS).

    This model finds the linear projections of two views that maximize their covariance while minimizing their residual variance.

    The objective function of PLS-ALS is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_i^Tw_i=1

    The algorithm alternates between updating :math:`w_1` and :math:`w_2` until convergence.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random seed for reproducibility, by default None
    epochs : int, optional
        Number of iterations to run the algorithm, by default 100
    deflation : str, optional
        Deflation scheme to use, by default "cca"
    initialization : str, optional
        Initialization scheme to use, by default "pls"
    tol : float, optional
        Tolerance for convergence, by default 1e-3
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        deflation="pls",
        accept_sparse=None,
        batch_size=None,
        dataloader_kwargs=None,
        epochs=100,
        val_split=None,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        callbacks=None,
        trainer_kwargs=None,
    ):
        super().__init__(
            latent_dimensions,
            copy_data,
            random_state,
            tol,
            deflation=deflation,
            accept_sparse=accept_sparse,
            batch_size=batch_size,
            dataloader_kwargs=dataloader_kwargs,
            epochs=epochs,
            val_split=val_split,
            learning_rate=learning_rate,
            initialization=initialization,
            callbacks=callbacks,
            trainer_kwargs=trainer_kwargs,
        )

    def _get_module(self, weights=None, k=None):
        return PlsAlsLoop(
            weights=weights,
            k=k,
        )

    def _more_tags(self):
        return {"multiview": True}


class PlsAlsLoop(BaseLoop):
    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        # Update each view using loop update function
        for view_index, view in enumerate(batch["views"]):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            self.weights[view_index] = np.cov(
                np.hstack((batch["views"][view_index], target[:, np.newaxis])).T
            )[:-1, -1]
            self.weights[view_index] /= np.linalg.norm(self.weights[view_index])



from abc import ABC
from typing import Iterable
import pytorch_lightning as pl
from tqdm import tqdm
import numpy as np
from cca_zoo.classical._iterative._base import BaseIterative


class BaseDeflation(BaseIterative, ABC):
    def _fit(self, views: Iterable[np.ndarray]):
        # if views is a tuple then convert to a list
        if isinstance(views, tuple):
            views = list(views)
        # tqdm for each latent dimension
        for k in tqdm(
            range(self.latent_dimensions), desc="Latent Dimension", leave=False
        ):
            train_dataloader, val_dataloader = self.get_dataloader(views)
            loop = self._get_module(weights=self.weights, k=k)
            # make a trainer
            trainer = pl.Trainer(
                max_epochs=self.epochs, callbacks=self.callbacks, **self.trainer_kwargs
            )
            trainer.fit(loop, train_dataloader, val_dataloader)
            # return the weights from the module. They will need to be changed from torch tensors to numpy arrays
            weights = loop.weights
            for i, (view, weight) in enumerate(zip(views, weights)):
                self.weights[i][:, k] = weight
                views[i] = self._deflate(view, weight)
            # if loop has tracked the objective, return the objective
            if hasattr(loop, "epoch_objective"):
                self.objective = loop.epoch_objective
        return self.weights

    def _deflate(self, residual: np.ndarray, weights: np.ndarray) -> np.ndarray:
        """Deflate view residual by CCA deflation.

        Parameters
        ----------
        residual : np.ndarray
            The current residual data matrix for a view
        weights : np.ndarray
            The current CCA weights for a view

        Returns
        -------
        np.ndarray
            The deflated residual data matrix for a view

        Raises
        ------
        ValueError
            If deflation method is not one of ["cca", "pls"]
        """
        # Compute the score vector for a view
        score = residual @ weights

        # Deflate the residual by different methods based on the deflation attribute
        if self.deflation == "cca":
            return residual - np.outer(score, score) @ residual / np.dot(score, score)
        elif self.deflation == "pls":
            return residual - np.outer(score, weights)
        else:
            raise ValueError(
                f"Invalid deflation method: {self.deflation}. "
                f"Must be one of ['cca', 'pls']."
            )



from sklearn.utils import check_random_state
from cca_zoo.classical import SCCA_PMD, ElasticCCA
from cca_zoo.classical import CCAEY

n = 50
rng = check_random_state(0)
X = rng.rand(n, 10)
Y = rng.rand(n, 11)
Z = rng.rand(n, 12)
# centre the data
X -= X.mean(axis=0)
Y -= Y.mean(axis=0)
Z -= Z.mean(axis=0)


def test_tracking():
    ey = CCAEY(latent_dimensions=2, track="loss").fit([X, Y])
    pmd = SCCA_PMD(latent_dimensions=2, track="weights_change", tau=0.5).fit([X, Y])
    elastic = ElasticCCA(latent_dimensions=2, track="loss").fit([X, Y])


def test_convergence():
    elastic = ElasticCCA(
        latent_dimensions=2,
        convergence_checking="loss",
        track="loss",
        alpha=5e-3,
        l1_ratio=0.5,
    ).fit([X, Y])
    pmd = SCCA_PMD(
        latent_dimensions=2,
        convergence_checking="weights_change",
        track="weights_change",
        tau=0.5,
    ).fit([X, Y])



from cca_zoo.classical import rCCA, MCCA
from typing import Iterable

import numpy as np


def reduce_dims(x):
    U, S, _ = np.linalg.svd(x, full_matrices=False)
    return U @ np.diag(S)


class PLSMixin:
    def total_correlation_(self, views: Iterable[np.ndarray], **kwargs) -> np.ndarray:
        """
        Returns the total correlation for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        correlation : numpy array of shape (n_views, latent_dimensions)

        """
        # Calculate total correlation in each view by SVD
        n = views[0].shape[0]
        views = list(views)
        for i, view in enumerate(views):
            if n < view.shape[1]:
                views[i] = reduce_dims(view)
        correlation = (
            MCCA(latent_dimensions=min([view.shape[1] for view in views]))
            .fit(views)
            .score(views)
            .sum()
        )
        return correlation

    def total_variance_(self, views: Iterable[np.ndarray], **kwargs) -> np.ndarray:
        """
        Returns the total variance for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        variance : numpy array of shape (n_views, latent_dimensions)

        """
        views = list(views)
        # Calculate total variance in each view by SVD
        n = views[0].shape[0]
        for i, view in enumerate(views):
            if n < view.shape[1]:
                views[i] = reduce_dims(view)
        variance = np.array(
            [np.sum(np.linalg.svd(view)[1] ** 2, keepdims=True) / n for view in views]
        )
        return variance

    def total_covariance_(self, views: Iterable[np.ndarray], **kwargs) -> float:
        """
        Returns the total covariance

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        covariance : float

        """
        views = list(views)
        # Calculate total covariance by SVD
        n = views[0].shape[0]
        for i, view in enumerate(views):
            if n < view.shape[1]:
                views[i] = reduce_dims(view)
        covariance = np.sum(np.linalg.svd(views[0].T @ views[1])[1]) / (n - 1)
        return covariance

    def explained_variance_(self, views: Iterable[np.ndarray], **kwargs) -> np.ndarray:
        """
        Returns the total variance for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        variance : numpy array of shape (n_views, latent_dimensions)

        """
        transformed_views = self.transform(views, **kwargs)
        variance = np.array([np.var(view, axis=0) for view in transformed_views])
        return variance

    def explained_covariance_(
        self, views: Iterable[np.ndarray], **kwargs
    ) -> np.ndarray:
        """
        Returns the total covariance for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        covariance : numpy array of shape (n_views, latent_dimensions)

        """
        transformed_views = self.transform(views, **kwargs)
        covariance = np.diag(
            np.cov(*transformed_views, rowvar=False)[
                : self.latent_dimensions, self.latent_dimensions :
            ]
        )
        return covariance

    def explained_covariance_ratio_(
        self, views: Iterable[np.ndarray], **kwargs
    ) -> np.ndarray:
        """
        Returns the explained covariance ratio for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        explained_covariance_ratio_ : numpy array of shape (n_views, latent_dimensions)

        """
        explained_covariance = self.explained_covariance_(views, **kwargs)
        covariance = self.total_covariance_(views, **kwargs)
        explained_covariance_ratio = explained_covariance / covariance
        return explained_covariance_ratio

    def explained_variance_ratio_(
        self, views: Iterable[np.ndarray], **kwargs
    ) -> np.ndarray:
        """
        Returns the explained variance ratio for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        explained_variance_ratio : numpy array of shape (n_views, latent_dimensions)

        """
        component_variance = self.explained_variance_(views, **kwargs)
        variance = self.total_variance_(views, **kwargs)
        explained_variance_ratio = component_variance / variance
        return explained_variance_ratio

    def explained_variance_cumulative_(
        self, views: Iterable[np.ndarray], **kwargs
    ) -> np.ndarray:
        """
        Returns the cumulative explained variance for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        explained_variance_cumulative_ : numpy array of shape (n_views, latent_dimensions)

        """
        explained_variance_ratio = self.explained_variance_ratio_(views, **kwargs)
        explained_variance_cumulative = np.cumsum(explained_variance_ratio, axis=1)
        return explained_variance_cumulative

    def explained_covariance_cumulative_(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the cumulative explained covariance for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        explained_covariance_cumulative_ : numpy array of shape (n_views, latent_dimensions)

        """
        explained_covariance_ratio = self.explained_covariance_ratio_(views, **kwargs)
        explained_covariance_cumulative = np.cumsum(explained_covariance_ratio)
        return explained_covariance_cumulative

    def total_variance_captured(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the total variance captured by the latent space

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        total_variance_captured : float

        """
        transformed_views = self.transform(views, **kwargs)
        total_variance_captured = self.total_variance_(transformed_views)
        return total_variance_captured

    def total_covariance_captured(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the total covariance captured by the latent space

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        total_covariance_captured : float

        """
        transformed_views = self.transform(views, **kwargs)
        total_covariance_captured = self.total_covariance_(transformed_views)
        return total_covariance_captured

    def total_correlation_captured(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the total correlation captured by the latent space

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        total_correlation_captured : float

        """
        transformed_views = self.transform(views, **kwargs)
        total_correlation_captured = self.total_correlation_(transformed_views)
        return total_correlation_captured


class PLS(rCCA, PLSMixin):
    r"""
    A class used to fit a simple PLS model. This model finds the linear projections of two views that maximize their covariance.

    Implements PLS by inheriting regularised CCA with maximal regularisation. This is equivalent to solving the following optimization problem:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_1^Tw_1=1

        w_2^Tw_2=1

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
    ):
        # Call the parent class constructor with c=1 to enable maximal regularization
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            c=1,
            random_state=random_state,
        )


class MPLS(MCCA, PLSMixin):
    r"""
    A class used to fit a mutiview PLS model. This model finds the linear projections of two views that maximize their covariance.

    Implements PLS by inheriting regularised CCA with maximal regularisation. This is equivalent to solving the following optimization problem:

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to use, by default 1
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
    ):
        # Call the parent class constructor with c=1 to enable maximal regularization
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            c=1,
            random_state=random_state,
        )



from cca_zoo.visualisation import plot_pairwise_scatter, plot_pairwise_correlations
import numpy as np
from cca_zoo.classical import CCA

x = np.random.rand(100, 10)
y = np.random.rand(100, 10)
x -= np.mean(x, axis=0)
y -= np.mean(y, axis=0)


def test_plotting():
    cca = CCA(latent_dimensions=3).fit([x, y])
    plot_pairwise_scatter(cca, [x, y])
    plot_pairwise_correlations(cca, [x, y])



from typing import Iterable

import torch
from torchmetrics import Metric

from cca_zoo.deep.objectives import _demean, inv_sqrtm


class MCCA(Metric):
    def __init__(self):
        super().__init__(dist_sync_on_step=False, compute_on_step=False)
        self.add_state("representations", default=[], persistent=False)

    def update(self, representations: Iterable[torch.Tensor] = None):
        for i, representation in enumerate(representations):
            if len(self.representations) < len(representations):
                self.representations.append([representation])
            else:
                self.representations[i].append(representation)
        self.latent_dims = representations[0].shape[1]

    @torch.no_grad()
    def compute(self):
        self.representations = [
            torch.cat(representation) for representation in self.representations
        ]
        return self.correlation(self.representations)

    def C(self, views):
        # Concatenate all views and from this get the cross-covariance matrix
        all_views = torch.cat(views, dim=1)
        C = torch.cov(all_views.T)
        C = C - torch.block_diag(*[torch.cov(view.T) for view in views])
        return C

    def D(self, views):
        # Get the block covariance matrix placing Xi^TX_i on the diagonal
        D = torch.block_diag(*[torch.cov(view.T) for i, view in enumerate(views)])
        return D

    def correlation(self, views):
        # Subtract the mean from each output
        views = _demean(views)

        C = self.C(views)
        D = self.D(views)

        R = inv_sqrtm(D, eps=1e-9)

        # In MCCA our eigenvalue problem Cv = lambda Dv
        C_whitened = R @ C @ R.T

        eigvals = torch.linalg.eigvalsh(C_whitened)

        # Sort eigenvalues so lowest first
        idx = torch.argsort(eigvals, descending=True)

        eigvals = eigvals[idx[: self.latent_dims]]

        return eigvals / 2



import torch

from ._dcca import DCCA


class DCCA_EY(DCCA):
    """

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A GeneralizedDeflation EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(self, latent_dimensions: int, encoders=None, r: float = 0, **kwargs):
        super().__init__(
            latent_dimensions=latent_dimensions, encoders=encoders, **kwargs
        )
        self.r = r

    def forward(self, views, **kwargs):
        z = []
        for i, encoder in enumerate(self.encoders):
            z.append(encoder(views[i]))  # encode each view into a latent representation
        return z  # return a list of latent representations

    def loss(self, views, **kwargs):
        # views here is a list of 'paired' views (i.e. [view1, view2])
        z = self(views)  # get the latent representations
        A, B = self.get_AB(z)  # get the cross-covariance and auto-covariance matrices
        rewards = 2 * torch.trace(
            A
        )  # compute the rewards as the sum of cross-covariances
        penalties = torch.trace(
            B @ B
        )  # compute the penalties as the squared Frobenius norm of auto-covariances
        return {
            "objective": -rewards + penalties,  # return the negative objective value
            "rewards": rewards,  # return the total rewards
            "penalties": penalties,  # return the penalties matrix
        }

    def get_AB(self, z):
        A = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the cross-covariance matrix
        B = torch.zeros(
            self.latent_dimensions, self.latent_dimensions, device=z[0].device
        )  # initialize the auto-covariance matrix
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += torch.cov(zi.T)  # add the auto-covariance of each view to B
                else:
                    A += torch.cov(torch.hstack((zi, zj)).T)[
                        self.latent_dimensions :, : self.latent_dimensions
                    ]  # add the cross-covariance of each pair of views to A
        return A / len(z), B / len(
            z
        )  # return the normalized matrices (divided by the number of views)



from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def plot_pairwise_correlations(
    model,
    views: Iterable[np.ndarray],
    ax=None,
    figsize=None,
    **kwargs,
):
    """
    Plots the pairwise correlations between the views in each dimension

    Parameters
    ----------
    views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
    ax : matplotlib axes object, optional
        If not provided, a new figure will be created.
    figsize : tuple, optional
        The size of the figure to create. If not provided, the default matplotlib figure size will be used.
    kwargs : any additional keyword arguments required by the given model

    Returns
    -------
    ax : matplotlib axes object

    """
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    corrs = model.pairwise_correlations(views, **kwargs)
    for i in range(corrs.shape[-1]):
        sns.heatmap(
            corrs[:, :, i],
            annot=True,
            vmin=-1,
            vmax=1,
            center=0,
            cmap="RdBu_r",
            ax=ax,
            **kwargs,
        )
        ax.set_title(f"Dimension {i + 1}")
    return ax


def plot_pairwise_scatter(
    model, views: Iterable[np.ndarray], ax=None, figsize=None, **kwargs
):
    """
    Plots the pairwise scatterplots between the views in each dimension

    Parameters
    ----------
    views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
    ax : matplotlib axes object, optional
        If not provided, a new figure will be created.
    figsize : tuple, optional
        The size of the figure to create. If not provided, the default matplotlib figure size will be used.
    kwargs : any additional keyword arguments required by the given model

    Returns
    -------
    ax : matplotlib axes object

    """
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    transformed_views = model.transform(views, **kwargs)
    for i in range(len(transformed_views)):
        for j in range(i + 1, len(transformed_views)):
            ax.scatter(transformed_views[i], transformed_views[j])
            ax.set_title(f"Dimension {i + 1} vs Dimension {j + 1}")
    return ax



from abc import abstractmethod
from typing import Iterable, List, Optional, Union, Any

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import Callback, EarlyStopping
from torch.utils import data
from torch.utils.data import DataLoader

from cca_zoo.data.deep import NumpyDataset
from cca_zoo.classical import MCCA, rCCA
from cca_zoo.classical._base import BaseModel
from cca_zoo.classical._dummy import DummyCCA

# Default Trainer kwargs
DEFAULT_TRAINER_KWARGS = dict(
    enable_checkpointing=False,
    logger=False,
    enable_model_summary=False,
    enable_progress_bar=False,
)

DEFAULT_LOADER_KWARGS = dict(
    num_workers=0, pin_memory=True, drop_last=False, shuffle=False
)


class BaseIterative(BaseModel):
    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        deflation="cca",
        accept_sparse=None,
        batch_size=None,
        dataloader_kwargs=None,
        epochs=1,
        val_split=None,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        callbacks: Optional[Union[List[Callback], Callback]] = None,
        trainer_kwargs=None,
        convergence_checking=None,
        patience=10,
        track=None,
        verbose=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            accept_sparse=accept_sparse,
        )
        self.tol = tol
        self.batch_size = batch_size
        self.epochs = epochs
        # validate the split
        if val_split is not None:
            if val_split <= 0 or val_split >= 1:
                raise ValueError("Validation split must be between 0 and 1")
        self.val_split = val_split
        self.learning_rate = learning_rate
        # validate the deflation method
        if deflation not in ["cca", "pls"]:
            raise ValueError("Deflation method must be one of ['cca','pls']")
        else:
            self.deflation = deflation
        # validate the initialization method
        if initialization not in ["random", "uniform", "pls", "cca"]:
            raise ValueError(
                "Initialization method must be one of ['random', 'uniform', 'pls', 'cca']"
            )
        else:
            self.initialization = initialization
        # validate the callbacks
        self.verbose = verbose
        self.patience = patience
        self.callbacks = callbacks or []
        # validate the convergence checking
        self.convergence_checking = convergence_checking
        # if convergence checking is a string
        if isinstance(self.convergence_checking, str):
            self.callbacks.append(
                ConvergenceCallback(
                    monitor=self.convergence_checking, min_delta=tol, patience=patience
                )
            )
        self.track = track
        if isinstance(self.track, str):
            self.callbacks.append(
                TrackingCallback(monitor=self.track, verbose=self.verbose)
            )
        self.dataloader_kwargs = dataloader_kwargs or DEFAULT_LOADER_KWARGS
        self.trainer_kwargs = trainer_kwargs or DEFAULT_TRAINER_KWARGS

    @abstractmethod
    def _get_module(self, weights=None, k=None):
        """Get the CCA module for training.

        Parameters
        ----------
        weights : list of np.ndarray, optional
            The initial weights for the CCA module, by default None
        k : int, optional
            The current component, by default None

        Returns
        -------
        pl.LightningModule
            The CCA module object
        """
        pass

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        self._validate_data(views)
        self._check_params()
        self._initialize(views)
        self.weights = self._fit(views)
        return self

    def _fit(self, views: Iterable[np.ndarray]):
        train_dataloader, val_dataloader = self.get_dataloader(views)
        loop = self._get_module(weights=self.weights)
        # make a trainer
        trainer = pl.Trainer(
            max_epochs=self.epochs,
            callbacks=self.callbacks,
            **self.trainer_kwargs,
        )
        trainer.fit(loop, train_dataloader, val_dataloader)
        # return the weights from the module. They will need to be changed from torch tensors to numpy arrays
        weights = loop.weights
        # if loop has tracked the objective, return the objective
        if hasattr(loop, "epoch_objective"):
            self.objective = loop.epoch_objective
        return weights

    def get_dataloader(self, views: Iterable[np.ndarray]):
        if self.batch_size is None:
            dataset = BatchNumpyDataset(views)
            collate_fn = lambda x: x[0]
        else:
            dataset = NumpyDataset(views)
            collate_fn = None
        if self.val_split is not None:
            train_size = int((1 - self.val_split) * len(dataset))
            val_size = len(dataset) - train_size
            dataset, val_dataset = data.random_split(dataset, [train_size, val_size])
            if self.batch_size is None:
                batch_size = val_size
            else:
                batch_size = self.batch_size
            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                **self.dataloader_kwargs,
                collate_fn=collate_fn,
            )
        else:
            val_loader = None
        if self.batch_size is None:
            batch_size = len(dataset)
        else:
            batch_size = self.batch_size
        train_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            **self.dataloader_kwargs,
            collate_fn=collate_fn,
        )

        return train_loader, val_loader

    def _initialize(self, views: Iterable[np.ndarray]):
        """Initialize the CCA weights using the initialization method or function.

        Parameters
        ----------
        views : Iterable[np.ndarray]
            The input views to initialize the CCA weights from
        """
        initializer = _default_initializer(
            self.initialization, self.random_state, self.latent_dimensions
        )
        # Fit the initializer on the input views and get the weights as numpy arrays
        self.weights = initializer.fit(views).weights
        self.weights = [weights.astype(np.float32) for weights in self.weights]


class BaseLoop(pl.LightningModule):
    def __init__(
        self,
        weights=None,
        k=None,
        automatic_optimization=False,
        tracking=False,
        convergence_checking=False,
    ):
        super().__init__()
        if k is not None:
            self.weights = [weight[:, k] for weight in weights]
        else:
            self.weights = weights
        self.automatic_optimization = automatic_optimization
        self.tracking = tracking
        self.convergence_checking = convergence_checking

    def objective(self, *args, **kwargs) -> float:
        raise NotImplementedError

    def forward(self, views: list) -> list:
        """Compute the score vectors for each view.

        Parameters
        ----------
        views : list
            The input views to compute the score vectors from

        Returns
        -------
        list
            The score vectors for each view
        """
        return [view @ weight for view, weight in zip(views, self.weights)]

    def configure_optimizers(self):
        """
        We don't need an optimizer for manual optimization.
        """
        pass


class BaseGradientLoop(BaseLoop):
    def __init__(
        self,
        weights: list = None,
        k: int = None,
        learning_rate: float = 1e-3,
        optimizer_kwargs: dict = None,
        tracking: bool = False,
        convergence_checking: bool = False,
    ):
        """Initialize the gradient-based CCA loop.

        Parameters
        ----------
        weights : list, optional
            The initial weights for the CCA loop, by default None
        k : int, optional
            The index of the latent dimension to use for the CCA loop, by default None
        learning_rate : float, optional
            The learning rate for the optimizer, by default 1e-3
        optimizer_kwargs : dict, optional
            The keyword arguments for the optimizer creation, by default None
        """
        super().__init__(
            weights=weights,
            k=k,
            automatic_optimization=True,
            tracking=tracking,
            convergence_checking=convergence_checking,
        )
        # Set the weights attribute as torch parameters with gradients
        self.weights = [
            torch.nn.Parameter(torch.from_numpy(weight), requires_grad=True)
            for weight in self.weights
        ]
        self.weights = torch.nn.ParameterList(self.weights)

        # Set the optimizer keyword arguments attribute with default values if none provided
        self.optimizer_kwargs = optimizer_kwargs or {}
        self.learning_rate = learning_rate

    def configure_optimizers(self):
        # construct optimizer using optimizer_kwargs
        optimizer_name = self.optimizer_kwargs.get("optimizer", "Adam")
        optimizer_kwargs = self.optimizer_kwargs.get("optimizer_kwargs", {})
        optimizer = getattr(torch.optim, optimizer_name)(
            self.weights, lr=self.learning_rate, **optimizer_kwargs
        )
        return optimizer

    def on_fit_end(self) -> None:
        # weights to numpy arrays from torch parameters
        weights = [weight.detach().cpu().numpy() for weight in self.weights]
        del self.weights
        self.weights = weights

    def forward(self, views):
        # if views are numpy arrays, convert to torch tensors
        if isinstance(views[0], np.ndarray):
            views = [torch.from_numpy(view) for view in views]
        return [view @ weight for view, weight in zip(views, self.weights)]


def _default_initializer(initialization, random_state, latent_dims):
    if initialization == "random":
        initializer = DummyCCA(latent_dims, random_state=random_state, uniform=False)
    elif initialization == "uniform":
        initializer = DummyCCA(latent_dims, random_state=random_state, uniform=True)
    elif initialization == "pls":
        initializer = rCCA(latent_dims, random_state=random_state, c=1)
    elif initialization == "cca":
        initializer = MCCA(latent_dims)
    else:
        raise ValueError(
            "Initialization {type} not supported. Pass a generator implementing this method"
        )
    return initializer


class ConvergenceCallback(EarlyStopping):
    def __init__(
        self,
        monitor: str = "loss",
        min_delta: float = 0.0,
        patience: int = 0,
        verbose: bool = False,
        mode: str = "min",
        strict: bool = True,
        check_finite: bool = True,
    ):
        super().__init__(
            monitor=monitor,
            min_delta=min_delta,
            patience=patience,
            verbose=verbose,
            mode=mode,
            strict=strict,
            check_finite=check_finite,
        )
        if monitor == "weights_change":
            self.stopping_threshold = min_delta

    def on_train_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs,
        batch: Any,
        batch_idx: int,
    ):
        pl_module.log(self.monitor, outputs[self.monitor].item())


class TrackingCallback(Callback):
    """
    Callback to track the objective function value during training
    """

    def __init__(self, monitor: str = "loss", verbose: bool = False):
        super().__init__()
        self.monitor = monitor
        self.verbose = verbose

    def on_train_start(self, trainer, pl_module):
        pl_module.batch_objective = []
        pl_module.epoch_objective = []

    def on_train_epoch_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        pl_module.batch_objective = []

    def on_train_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs,
        batch: Any,
        batch_idx: int,
    ):
        pl_module.batch_objective.append(outputs[self.monitor].item())
        if self.verbose:
            # Print the objective function value and the current batch index
            print(
                f"Objective: {outputs[self.monitor].item():.3f} | Batch: {batch_idx}",
                end="\r",
            )

    def on_train_epoch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        # epoch objective values are the mean of the batch objective values
        pl_module.epoch_objective.append(np.mean(pl_module.batch_objective))
        if self.verbose:
            # Print a new line after the last batch
            print()
            print(
                f"Objective: {pl_module.epoch_objective[-1]:.3f} | Epoch: {trainer.current_epoch}",
            )


class BatchNumpyDataset:
    def __init__(self, views, labels=None):
        self.views = [view.astype(np.float32) for view in views]
        self.labels = labels if labels is not None else None

    def __len__(self):
        return 1

    def __getitem__(self, index):
        if self.labels is not None:
            return {"views": self.views, "label": self.labels}
        else:
            return {"views": self.views}



from ._altmaxvar import AltMaxVar
from ._elasticnet import SCCA_IPLS, ElasticCCA
from ._ey import CCAEY, PLSEY
from ._gh import CCAGH
from ._pls_als import PLS_ALS
from ._pmd import SCCA_PMD
from ._scca_parkhomenko import SCCA_Parkhomenko
from ._scca_span import SCCA_Span
from ._stochasticpls import PLSStochasticPower
from ._svd import CCASVD, PLSSVD

__all__ = [
    "ElasticCCA",
    "SCCA_IPLS",
    "PLS_ALS",
    "SCCA_PMD",
    # "SCCA_ADMM",
    "SCCA_Parkhomenko",
    "SCCA_Span",
    # "SWCCA",
    # "GradKCCA",
    # "SCCA_HSIC",
    "AltMaxVar",
    "CCAEY",
    "PLSEY",
    "CCAGH",
    "CCASVD",
    "PLSSVD",
    # "PLSGHA",
]



from typing import Union

import torch

from cca_zoo.classical._iterative._ey import CCAEY, EYLoop
from cca_zoo.classical._pls import PLSMixin


class CCASVD(CCAEY):
    def _get_module(self, weights=None, k=None):
        return SVDLoop(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
            objective="cca",
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class PLSSVD(CCASVD, PLSMixin):
    def _get_module(self, weights=None, k=None):
        return SVDLoop(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
            objective="pls",
            tracking=self.track,
            convergence_checking=self.convergence_checking,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class SVDLoop(EYLoop):
    def loss(self, views, views2=None, **kwargs):
        z = self(views)
        C = torch.cov(torch.hstack(z).T)
        latent_dims = z[0].shape[1]

        Cxy = C[:latent_dims, latent_dims:]
        Cxx = C[:latent_dims, :latent_dims]

        if views2 is None:
            Cyy = C[latent_dims:, latent_dims:]
        else:
            z2 = self(views2)
            Cyy = torch.cov(torch.hstack(z2).T)[latent_dims:, latent_dims:]

        rewards = torch.trace(2 * Cxy)
        penalties = torch.trace(Cxx @ Cyy)

        return {
            "loss": -rewards + penalties,
            "rewards": rewards,
            "penalties": penalties,
        }



"""
Test the kernel methods
"""


import numpy as np
from cca_zoo.classical import KCCA, KTCCA, KGCCA, MCCA, GCCA, TCCA

np.random.seed(1)
X = np.random.normal(size=(100, 10))
Y = np.random.normal(size=(100, 10))
Z = np.random.normal(size=(100, 10))


# test that MCCA is the same as KCCA with linear kernel
def test_MCCA_KCCA():
    kcca = KCCA(latent_dimensions=2)
    mcca = MCCA(latent_dimensions=2)
    kcca.fit([X, Y, Z])
    mcca.fit([X, Y, Z])
    mcca_score = mcca.score([X, Y, Z])
    kcca_score = kcca.score([X, Y, Z])
    assert np.allclose(mcca_score, kcca_score)


# test that GCCA is the same as KGCCA with linear kernel
def test_GCCA_KGCCA():
    kgcca = KGCCA(latent_dimensions=2)
    gcca = GCCA(latent_dimensions=2)
    kgcca.fit([X, Y, Z])
    gcca.fit([X, Y, Z])
    gcca_score = gcca.score([X, Y, Z])
    kgcca_score = kgcca.score([X, Y, Z])
    assert np.allclose(gcca_score, kgcca_score)


# test that TCCA is the same as KTCCA with linear kernel
def test_TCCA_KTCCA():
    ktcca = KTCCA(latent_dimensions=2)
    tcca = TCCA(latent_dimensions=2)
    ktcca.fit([X, Y, Z])
    tcca.fit([X, Y, Z])
    tcca_score = tcca.score([X, Y, Z])
    ktcca_score = ktcca.score([X, Y, Z])
    assert np.allclose(tcca_score, ktcca_score)


def test_rbf_kernel():
    kcca = KCCA(latent_dimensions=2, kernel="rbf").fit((X, Y, Z))
    kgcca = KGCCA(latent_dimensions=2, kernel="rbf").fit((X, Y, Z))
    ktcca = KTCCA(latent_dimensions=2, kernel="rbf").fit((X, Y, Z))


def test_poly_kernel():
    kcca = KCCA(latent_dimensions=2, kernel="poly").fit((X, Y, Z))
    kgcca = KGCCA(latent_dimensions=2, kernel="poly").fit((X, Y, Z))
    ktcca = KTCCA(latent_dimensions=2, kernel="poly").fit((X, Y, Z))


def test_sigmoid_kernel():
    kcca = KCCA(latent_dimensions=2, kernel="sigmoid").fit((X, Y, Z))
    kgcca = KGCCA(latent_dimensions=2, kernel="sigmoid").fit((X, Y, Z))
    ktcca = KTCCA(latent_dimensions=2, kernel="sigmoid").fit((X, Y, Z))


def test_cosine_kernel():
    kcca = KCCA(latent_dimensions=2, kernel="cosine").fit((X, Y, Z))
    kgcca = KGCCA(latent_dimensions=2, kernel="cosine").fit((X, Y, Z))
    ktcca = KTCCA(latent_dimensions=2, kernel="cosine").fit((X, Y, Z))


def test_callable_kernel():
    def my_kernel(X, Y, **kwargs):
        return np.dot(X, Y.T)

    kcca = KCCA(latent_dimensions=2, kernel=my_kernel).fit((X, Y, Z))
    kgcca = KGCCA(latent_dimensions=2, kernel=my_kernel).fit((X, Y, Z))
    ktcca = KTCCA(latent_dimensions=2, kernel=my_kernel).fit((X, Y, Z))



from typing import Iterable, Union

import numpy as np
import torch
from skprox.proximal_operators import _proximal_operators

from cca_zoo.classical._iterative._base import BaseIterative, BaseLoop
from cca_zoo.utils import _process_parameter


class AltMaxVar(BaseIterative):
    def __init__(
        self,
        latent_dimensions=1,
        copy_data=True,
        random_state=None,
        epochs=100,
        tol=1e-3,
        proximal="L1",
        positive=False,
        tau: Union[Iterable[float], float] = None,
        proximal_params: Iterable[dict] = None,
        gamma=0.1,
        learning_rate=0.1,
        T=100,
        convergence_checking=None,
        track=None,
        verbose=False,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            epochs=epochs,
            convergence_checking=convergence_checking,
            track=track,
            verbose=verbose,
        )
        self.tau = tau
        self.proximal = proximal
        self.proximal_params = proximal_params
        self.gamma = gamma
        self.learning_rate = learning_rate
        self.T = T
        self.positive = positive

    def _get_module(self, weights=None, k=None):
        return AltMaxVarLoop(
            weights=weights,
            k=k,
            gamma=self.gamma,
            T=self.T,
            learning_rate=self.learning_rate,
            proximal_operators=self.proximal_operators,
        )

    def _check_params(self):
        self.proximal = _process_parameter(
            "proximal", self.proximal, "L1", self.n_views_
        )
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )
        self.tau = _process_parameter("tau", self.tau, 0, self.n_views_)
        self.sigma = self.tau
        self.proximal_operators = [
            self._get_proximal(view) for view in range(self.n_views_)
        ]

    def _get_proximal(self, view):
        if callable(self.proximal[view]):
            params = self.proximal_params[view] or {}
        else:
            params = {
                "sigma": self.sigma[view],
                "positive": self.positive[view],
            }
        return _proximal_operators(self.proximal[view], **params)

    def _more_tags(self):
        return {"multiview": True}


class AltMaxVarLoop(BaseLoop):
    def __init__(
        self,
        weights,
        k=None,
        gamma=0.1,
        T=100,
        learning_rate=1e-1,
        proximal_operators=None,
    ):
        super().__init__(weights, k)
        self.gamma = gamma
        self.proximal_operators = proximal_operators
        self.T = T
        self.learning_rate = learning_rate

    def _get_target(self, scores):
        if hasattr(self, "G"):
            R = self.gamma * scores.mean(axis=0) + (1 - self.gamma) * self.G
        else:
            R = scores.mean(axis=0)
        U, S, Vt = np.linalg.svd(R, full_matrices=False)
        G = U @ Vt
        return G

    def objective(self, views, scores, weights) -> int:
        least_squares = (np.linalg.norm(scores - self.G, axis=(1, 2)) ** 2).sum()
        regularization = np.array(
            [
                self.proximal_operators[view](weights[view])
                for view in range(self.n_views)
            ]
        ).sum()
        return least_squares + regularization

    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        old_weights = self.weights.copy()
        self.G = self._get_target(scores)
        converged = False
        for i, view in enumerate(batch["views"]):
            t = 0
            # initialize the previous weights to None
            prev_weights = None
            while t < self.T and not converged:
                # update the weights using the gradient descent and proximal operator
                self.weights[i] -= self.learning_rate * (
                    view.T @ (view @ self.weights[i] - self.G)
                )
                self.weights[i] = self.proximal_operators[i].prox(
                    self.weights[i], self.learning_rate
                )
                # check if the weights have changed significantly from the previous iteration
                if prev_weights is not None and np.allclose(
                    self.weights[i], prev_weights
                ):
                    # if yes, set converged to True and break the loop
                    converged = True
                    break
                # update the previous weights for the next iteration
                prev_weights = self.weights[i]
                t += 1

        # if tracking or convergence_checking is enabled, compute the objective function
        if self.tracking or self.convergence_checking:
            objective = self.objective(batch["views"])
            # check that the maximum change in weights is smaller than the tolerance times the maximum absolute value of the weights
            weights_change = torch.tensor(
                np.max(
                    [
                        np.max(np.abs(old_weights[i] - self.weights[i]))
                        / np.max(np.abs(self.weights[i]))
                        for i in range(len(self.weights))
                    ]
                )
            )
            return {"loss": torch.tensor(objective), "weights_change": weights_change}



import itertools
from abc import abstractmethod
from typing import Iterable, Union

import numpy as np
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted, check_random_state


class BaseModel(BaseEstimator, MultiOutputMixin, RegressorMixin):
    """
    A base class for multivariate latent variable classical.

    This class implements common methods and attributes for fitting and transforming
    multiple views of data using latent variable classical. It inherits from scikit-learn's
    BaseEstimator, MultiOutputMixin and RegressorMixin classes.

    Parameters
    ----------
    latent_dimensions : int, optional
        Number of latent dimensions to fit. Default is 1.
    copy_data : bool, optional
        Whether to copy the data. Default is True.
    accept_sparse : bool, optional
        Whether to accept sparse data. Default is False.
    random_state : int, RandomState instance or None, optional (default=None)
        Pass an int for reproducible output across multiple function calls.

    Attributes
    ----------
    n_views_ : int
        Number of views.
    n_features_ : list of int
        Number of features for each view.

    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        accept_sparse=False,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        self.latent_dimensions = latent_dimensions
        self.copy_data = copy_data
        self.accept_sparse = accept_sparse
        self.random_state = check_random_state(random_state)
        self.dtypes = FLOAT_DTYPES

    def _validate_data(
        self,
        views: Iterable[np.ndarray],
    ):
        if not all(view.shape[0] == views[0].shape[0] for view in views):
            raise ValueError("All views must have the same number of samples")
        if not all(view.ndim == 2 for view in views):
            raise ValueError("All views must have 2 dimensions")
        if not all(view.dtype in self.dtypes for view in views):
            raise ValueError("All views must have dtype of {}.".format(self.dtypes))
        if not all(view.shape[1] >= self.latent_dimensions for view in views):
            raise ValueError(
                "All views must have at least {} features.".format(
                    self.latent_dimensions
                )
            )
        self.n_views_ = len(views)
        self.n_features_ = [view.shape[1] for view in views]
        self.n_samples_ = views[0].shape[0]

    def _check_params(self):
        """
        Checks the parameters of the model.
        """
        pass

    @abstractmethod
    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        """
        Fits the model to the given data

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs: any additional keyword arguments required by the given model

        Returns
        -------
        self : object

        """
        return self

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        """

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        check_is_fitted(self, attributes=["weights"])
        transformed_views = []
        for i, view in enumerate(views):
            transformed_view = view @ self.weights[i]
            transformed_views.append(transformed_view)
        return transformed_views

    def fit_transform(self, views: Iterable[np.ndarray], **kwargs):
        """
        Fits the model to the given data and returns the transformed views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        return self.fit(views, **kwargs).transform(views, **kwargs)

    def pairwise_correlations(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the pairwise correlations between the views in each dimension

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        pairwise_correlations : numpy array of shape (n_views, n_views, latent_dimensions)

        """
        transformed_views = self.transform(views, **kwargs)
        all_corrs = []
        for x, y in itertools.product(transformed_views, repeat=2):
            all_corrs.append(
                np.diag(
                    np.corrcoef(x.T, y.T)[
                        : self.latent_dimensions, self.latent_dimensions :
                    ]
                )
            )
        all_corrs = np.array(all_corrs).reshape(
            (self.n_views_, self.n_views_, self.latent_dimensions)
        )
        return all_corrs

    def score(self, views: Iterable[np.ndarray], y=None, **kwargs):
        """
        Returns the average pairwise correlation between the views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        score : float


        """
        # by default return the average pairwise correlation in each dimension (for 2 views just the correlation)
        pair_corrs = self.pairwise_correlations(views, **kwargs)
        # sum all the pairwise correlations for each dimension. Subtract the self correlations. Divide by the number of views. Gives average correlation
        dim_corrs = np.sum(pair_corrs, axis=(0, 1)) - pair_corrs.shape[0]
        # number of pairs is n_views choose 2
        num_pairs = (self.n_views_ * (self.n_views_ - 1)) / 2
        dim_corrs = dim_corrs / (2 * num_pairs)
        return dim_corrs

    def factor_loadings(self, views: Iterable[np.ndarray], normalize=True, **kwargs):
        """
        Returns the factor loadings for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        normalize : bool, optional
            Whether to normalize the factor loadings. Default is True.
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        factor_loadings : list of numpy arrays

        """
        transformed_views = self.transform(views, **kwargs)
        if normalize:
            loadings = [
                np.corrcoef(view, transformed_view, rowvar=False)[
                    : view.shape[1], view.shape[1] :
                ]
                for view, transformed_view in zip(views, transformed_views)
            ]
        else:
            loadings = [
                view.T @ transformed_view
                for view, transformed_view in zip(views, transformed_views)
            ]
        return loadings



from typing import Iterable, Union

import numpy as np

from cca_zoo.classical._iterative._base import BaseIterative, BaseLoop
from cca_zoo.classical._search import _delta_search, support_threshold
from cca_zoo.utils import _process_parameter


class SCCA_Span(BaseIterative):
    r"""
    Fits a Sparse CCA model using SpanCCA.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \text{l1_ratio}\|w_i\|_1\}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

    References
    ----------
    Asteris, Megasthenis, et al. "A simple and provable algorithm for sparse diagonal CCA." International Conference on Machine Learning. PMLR, 2016.


    Examples
    --------
    >>> from cca_zoo.classical import SCCA_Span
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_Span(regularisation="l0", tau=[2, 2])
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.84556666])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        epochs: int = 100,
        copy_data=True,
        initialization: str = "uniform",
        tol: float = 1e-3,
        regularisation="l0",
        tau: Union[Iterable[Union[float, int]], Union[float, int]] = None,
        rank=1,
        positive: Union[Iterable[bool], bool] = None,
        random_state=None,
        deflation="cca",
        verbose=0,
    ):
        super().__init__(
            latent_dimensions=latent_dimensions,
            epochs=epochs,
            copy_data=copy_data,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            verbose=verbose,
        )
        self.tau = tau
        self.regularisation = regularisation
        self.rank = rank
        self.positive = positive

    def _check_params(self):
        """check number of views=2"""
        if self.n_views_ != 2:
            raise ValueError(f"SCCA_Span requires only 2 views")
        self.max_obj = 0
        if self.regularisation == "l0":
            self.update = support_threshold
        elif self.regularisation == "l1":
            self.update = _delta_search
        self.tau = _process_parameter("tau", self.tau, 0, self.n_views_)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views_
        )

    def _get_module(self, weights=None, k=None):
        return SpanLoop(
            weights=weights,
            k=k,
            regularisation=self.regularisation,
            tau=self.tau,
            positive=self.positive,
            rank=self.rank,
            random_state=self.random_state,
        )


class SpanLoop(BaseLoop):
    def __init__(
        self,
        weights,
        k=None,
        regularisation="l0",
        tau=None,
        positive=False,
        rank=1,
        random_state=None,
    ):
        super().__init__(weights=weights, k=k)
        self.regularisation = regularisation
        self.tau = tau
        self.positive = positive
        if self.regularisation == "l0":
            self.update = support_threshold
        elif self.regularisation == "l1":
            self.update = _delta_search
        self.rank = rank
        self.random_state = random_state

    def training_step(self, batch, batch_idx):
        # if P, D, Q not initialised, initialise them
        if getattr(self, "P", None) is None:
            self._initialize(batch["views"])
        c = self.random_state.randn(self.rank)
        c /= np.linalg.norm(c)
        a = self.P @ np.diag(self.D) @ c
        u = self.update(a, self.tau[0])
        u /= np.linalg.norm(u)
        b = self.Q @ np.diag(self.D) @ self.P.T @ u
        v = self.update(b, self.tau[1])
        v /= np.linalg.norm(v)
        if b.T @ v > self.max_obj:
            self.max_obj = b.T @ v
            self.weights[0] = u
            self.weights[1] = v

    def _initialize(self, views):
        self.max_obj = 0
        cov = views[0].T @ views[1] / views[0].shape[0]
        # Perform SVD on im and obtain individual matrices
        P, D, Q = np.linalg.svd(cov, full_matrices=True)
        self.P = P[:, : self.rank]
        self.D = D[: self.rank]
        self.Q = Q[: self.rank, :].T



import numpy as np
import scipy.sparse as sp
from sklearn.utils.validation import check_random_state

from cca_zoo.classical import (
    CCA,
    GCCA,
    KCCA,
    KGCCA,
    KTCCA,
    MCCA,
    PLS,
    PLS_ALS,
    TCCA,
    rCCA,
    PCACCA,
)

n = 50
rng = check_random_state(0)
X = rng.rand(n, 11)
Y = rng.rand(n, 10)
Z = rng.rand(n, 12)
X_sp = sp.random(n, 10, density=0.5, random_state=rng)
Y_sp = sp.random(n, 11, density=0.5, random_state=rng)
# centre the data
X -= X.mean(axis=0)
Y -= Y.mean(axis=0)
Z -= Z.mean(axis=0)
X_sp -= X_sp.mean(axis=0)
Y_sp -= Y_sp.mean(axis=0)


def test_unregularized_methods():
    # This function tests unregularized CCA methods. The idea is that all of these should give the same result.
    latent_dims = 2
    rcca = rCCA(latent_dimensions=latent_dims).fit([X, Y])
    cca = CCA(latent_dimensions=latent_dims).fit([X, Y])
    gcca = GCCA(latent_dimensions=latent_dims).fit([X, Y])
    mcca = MCCA(latent_dimensions=latent_dims, pca=False).fit([X, Y])
    mcca_pca = MCCA(latent_dimensions=latent_dims, pca=True).fit([X, Y])
    kcca = KCCA(latent_dimensions=latent_dims).fit([X, Y])
    kgcca = KGCCA(latent_dimensions=latent_dims).fit([X, Y])
    tcca = TCCA(latent_dimensions=latent_dims).fit([X, Y])
    pcacca = PCACCA(latent_dimensions=latent_dims).fit([X, Y])

    # Get the correlation scores for each method
    corr_rcca = rcca.score((X, Y))
    corr_cca = cca.score((X, Y))
    corr_gcca = gcca.score((X, Y))
    corr_mcca = mcca.score((X, Y))
    corr_mcca_pca = mcca_pca.score((X, Y))
    corr_pcacca = pcacca.score((X, Y))
    corr_kcca = kcca.score((X, Y))
    corr_kgcca = kgcca.score((X, Y))
    corr_tcca = tcca.score((X, Y))

    # Assert that the correlation scores are all equal
    assert np.testing.assert_array_almost_equal(corr_cca, corr_mcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_gcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_kcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_tcca, decimal=1) is None
    assert (
        np.testing.assert_array_almost_equal(corr_kgcca, corr_gcca, decimal=1) is None
    )


def test_unregularized_multi():
    # This function tests unregularized CCA methods for more than 2 views. The idea is that all of these should give the same result.
    latent_dims = 2
    gcca = GCCA(latent_dimensions=latent_dims).fit((X, Y, Z))
    mcca = MCCA(latent_dimensions=latent_dims).fit((X, Y, Z))
    kcca = KCCA(latent_dimensions=latent_dims).fit((X, Y, Z))
    # Get the correlation scores for each method
    corr_gcca = gcca.score((X, Y, Z))
    corr_mcca = mcca.score((X, Y, Z))
    corr_kcca = kcca.score((X, Y, Z))

    # Assert that the correlation scores are all equal
    assert np.testing.assert_array_almost_equal(corr_mcca, corr_gcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_mcca, corr_kcca, decimal=1) is None
    # Get the correlation scores for each method
    corr_gcca = gcca.score((X, Y, Z))
    corr_mcca = mcca.score((X, Y, Z))
    corr_kcca = kcca.score((X, Y, Z))

    # Assert that the correlation scores are all equal
    assert np.testing.assert_array_almost_equal(corr_mcca, corr_gcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_mcca, corr_kcca, decimal=1) is None


def test_pls():
    # This function tests PLS and PLS_ALS
    pls_als = PLS_ALS(latent_dimensions=3, random_state=0)
    pls = PLS(latent_dimensions=3)

    # Fit both classical to the same data
    pls_als.fit([X, Y])
    pls.fit((X, Y))

    # Assert that the scores are close
    pls_score = pls.score((X, Y))
    pls_als_score = pls_als.score((X, Y))

    assert np.allclose(np.abs(pls_als_score), pls_score, rtol=1e-1)


def test_TCCA():
    # This function tests TCCA and KTCCA
    latent_dims = 1
    tcca = TCCA(latent_dimensions=latent_dims, c=[0.2, 0.2, 0.2]).fit([X, X, Y])
    ktcca = KTCCA(latent_dimensions=latent_dims, c=[0.2, 0.2]).fit([X, Y])

    # Get the correlation scores for each method
    corr_tcca = tcca.score((X, X, Y))
    corr_ktcca = ktcca.score((X, Y))

    # Assert that the correlation scores are greater than 0.1
    assert corr_tcca > 0.1
    assert corr_ktcca > 0.1



import numpy as np
import scipy.sparse as sp
from sklearn.utils import check_random_state

from cca_zoo.model_selection import (
    cross_validate,
    learning_curve,
    permutation_test_score,
)
from cca_zoo.classical import MPLS

n = 100
rng = check_random_state(0)
X_low = rng.rand(n, 5)
Y_low = rng.rand(n, 5)
n = 5
X_high = rng.rand(n, 100)
Y_high = rng.rand(n, 101)
# centre the data
X_low -= X_low.mean(axis=0)
Y_low -= Y_low.mean(axis=0)
X_high -= X_high.mean(axis=0)
Y_high -= Y_high.mean(axis=0)


def test_explained_variance():
    for X in [X_low, X_high]:
        # Test that explained variance is between 0 and 1
        pls = MPLS(latent_dimensions=5).fit((X, X))
        explained_variance = pls.explained_variance_((X, X))
        explained_variance_ratio = pls.explained_variance_ratio_((X, X))
        explained_variance_cumulative = pls.explained_variance_cumulative_((X, X))
        # explained_variance_ratio should sum to 1 for each view
        assert np.allclose(explained_variance_ratio.sum(axis=1), 1)
        # explained_variance_cumulative should be monotonically increasing
        assert np.all(np.diff(explained_variance_cumulative, axis=1) >= 0)


def test_explained_covariance():
    for X in [X_low, X_high]:
        # Test that explained covariance is between 0 and 1
        pls = MPLS(latent_dimensions=5).fit((X, X))
        explained_covariance = pls.explained_covariance_((X, X))
        explained_covariance_ratio = pls.explained_covariance_ratio_((X, X))
        explained_covariance_cumulative = pls.explained_covariance_cumulative_((X, X))
        # explained_covariance_ratio should sum to 1 for each view
        assert np.allclose(explained_covariance_ratio.sum(), 1)
        # explained_covariance_cumulative should be monotonically increasing
        assert np.all(np.diff(explained_covariance_cumulative) >= 0)


# def test_explained_correlation():


def test_validation():
    # Test that validation works
    pls = MPLS(latent_dimensions=1).fit((X_low, Y_low))
    cross_validate(pls, (X_low, Y_low))
    permutation_test_score(pls, (X_low, Y_low))
    learning_curve(pls, (X_low, Y_low))



import numpy as np
import pytorch_lightning as pl
from sklearn.utils.validation import check_random_state
from torch import manual_seed
from torch.utils.data import random_split

from cca_zoo.data.deep import NumpyDataset, check_dataset, get_dataloaders
from cca_zoo.deep import (
    DCCA,
    DCCA_EY,
    DCCA_GH,
    DCCA_SVD,
    DCCA_NOI,
    DCCA_SDL,
    DCCAE,
    DTCCA,
    DVCCA,
    DGCCA,
    BarlowTwins,
    SplitAE,
    architectures,
    objectives,
)
from cca_zoo.classical import CCA, MCCA, GCCA

manual_seed(0)
rng = check_random_state(0)
X = rng.rand(256, 10)
Y = rng.rand(256, 12)
Z = rng.rand(256, 14)
X -= X.mean(axis=0)
Y -= Y.mean(axis=0)
Z -= Z.mean(axis=0)
X_conv = rng.rand(256, 1, 16, 16)
Y_conv = rng.rand(256, 1, 16, 16)
X_conv -= X_conv.mean(axis=0)
Y_conv -= Y_conv.mean(axis=0)
dataset = NumpyDataset([X, Y, Z])
check_dataset(dataset)
train_dataset, val_dataset = random_split(dataset, [256 - 56, 56])
loader = get_dataloaders(dataset)
train_loader, val_loader = get_dataloaders(train_dataset, val_dataset)
conv_dataset = NumpyDataset((X_conv, Y_conv))
conv_loader = get_dataloaders(conv_dataset)
train_ids = train_dataset.indices
trainer_kwargs = dict(
    enable_checkpointing=False,
    logger=False,
    enable_model_summary=False,
    enable_progress_bar=False,
)


def test_numpy_dataset():
    dataset = NumpyDataset([X, Y, Z])
    check_dataset(dataset)
    get_dataloaders(dataset)


def test_linear_mcca():
    max_epochs = 100
    latent_dimensions = 2
    mcca = MCCA(latent_dimensions=latent_dimensions).fit((X, Y, Z))
    # DCCA_MCCA
    encoder_1 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    encoder_3 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=14
    )
    dmcca = DCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2, encoder_3],
        lr=1e-1,
        objective=objectives.MCCA,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dmcca, loader)
    assert (
        np.testing.assert_array_almost_equal(
            mcca.score((X, Y, Z)).sum(), dmcca.score(loader).sum(), decimal=2
        )
        is None
    )


def test_linear_gcca():
    max_epochs = 100
    latent_dimensions = 2
    gcca = GCCA(latent_dimensions=latent_dimensions).fit((X, Y, Z))
    # DCCA_GCCA
    encoder_1 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    encoder_3 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=14
    )
    dgcca = DGCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2, encoder_3],
        lr=1e-2,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dgcca, loader)
    assert (
        np.testing.assert_array_almost_equal(
            gcca.score((X, Y, Z)).sum(), dgcca.score(loader).sum(), decimal=2
        )
        is None
    )


def test_DTCCA_methods():
    max_epochs = 100
    # check that DTCCA is equivalent to CCA for 2 views with linear encoders
    latent_dimensions = 2
    cca = CCA(latent_dimensions=latent_dimensions)
    encoder_1 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.LinearEncoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dtcca = DTCCA(
        latent_dimensions=latent_dimensions, encoders=[encoder_1, encoder_2], lr=1e-2
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dtcca, train_loader)
    z = dtcca.transform(train_loader)
    assert (
        np.testing.assert_array_almost_equal(
            cca.fit((X[train_ids], Y[train_ids]))
            .score((X[train_ids], Y[train_ids]))
            .sum(),
            cca.fit((z)).score((z)).sum(),
            decimal=1,
        )
        is None
    )


def test_DCCA_methods():
    N = len(train_dataset)
    max_epochs = 100
    latent_dimensions = 2
    cca = CCA(latent_dimensions=latent_dimensions).fit((X, Y))
    # DCCA
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dcca = DCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        objective=objectives.CCA,
        lr=1e-3,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dcca, train_loader, val_dataloaders=val_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dcca.score(train_loader).sum()
        )
        is None
    )
    # DCCA_GH
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dcca_gh = DCCA_GH(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        lr=1e-1,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dcca_gh, train_loader, val_dataloaders=val_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dcca_gh.score(train_loader).sum()
        )
        is None
    )
    # DCCA_SVD
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dcca_svd = DCCA_SVD(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        lr=1e-1,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dcca_svd, train_loader, val_dataloaders=val_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dcca_svd.score(train_loader).sum()
        )
        is None
    )
    # DCCA_EY
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dcca_ey = DCCA_EY(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        lr=1e-1,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dcca_ey, train_loader, val_dataloaders=val_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dcca_ey.score(train_loader).sum()
        )
        is None
    )
    # DCCA_NOI
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dcca_noi = DCCA_NOI(latent_dimensions, N, encoders=[encoder_1, encoder_2], rho=0.2)
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dcca_noi, train_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dcca_noi.score(train_loader).sum()
        )
        is None
    )
    # Soft Decorrelation (_stochastic Decorrelation Loss)
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    sdl = DCCA_SDL(
        latent_dimensions, N, encoders=[encoder_1, encoder_2], lam=1e-2, lr=1e-3
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(sdl, train_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), sdl.score(train_loader).sum()
        )
        is None
    )
    # Barlow Twins
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    barlowtwins = BarlowTwins(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(barlowtwins, train_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), barlowtwins.score(train_loader).sum()
        )
        is None
    )
    # DGCCA
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dgcca = DCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        objective=objectives.GCCA,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dgcca, train_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dgcca.score(train_loader).sum()
        )
        is None
    )
    # DMCCA
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dmcca = DCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        objective=objectives.MCCA,
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dmcca, train_loader)
    assert (
        np.testing.assert_array_less(
            cca.score((X, Y)).sum(), dmcca.score(train_loader).sum()
        )
        is None
    )


def test_DCCAE_methods():
    max_epochs = 2
    latent_dimensions = 2
    encoder_1 = architectures.CNNEncoder(
        latent_dimensions=latent_dimensions, feature_size=(16, 16)
    )
    encoder_2 = architectures.CNNEncoder(
        latent_dimensions=latent_dimensions, feature_size=(16, 16)
    )
    decoder_1 = architectures.CNNDecoder(
        latent_dimensions=latent_dimensions, feature_size=(16, 16)
    )
    decoder_2 = architectures.CNNDecoder(
        latent_dimensions=latent_dimensions, feature_size=(16, 16)
    )
    # SplitAE
    splitae = SplitAE(
        latent_dimensions=latent_dimensions,
        encoder=encoder_1,
        decoders=[decoder_1, decoder_2],
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(splitae, conv_loader)
    # DCCAE
    dccae = DCCAE(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        decoders=[decoder_1, decoder_2],
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dccae, conv_loader)


def test_DVCCA_p_methods():
    max_epochs = 2
    latent_dimensions = 2
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10, variational=True
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12, variational=True
    )
    private_encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10, variational=True
    )
    private_encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12, variational=True
    )
    decoder_1 = architectures.Decoder(
        latent_dimensions=2 * latent_dimensions, feature_size=10
    )
    decoder_2 = architectures.Decoder(
        latent_dimensions=2 * latent_dimensions, feature_size=12
    )
    # DVCCA
    dvcca = DVCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        decoders=[decoder_1, decoder_2],
        private_encoders=[private_encoder_1, private_encoder_2],
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dvcca, train_loader)
    dvcca.transform(train_loader)


def test_DVCCA_methods():
    max_epochs = 2
    latent_dimensions = 2
    encoder_1 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=10, variational=True
    )
    encoder_2 = architectures.Encoder(
        latent_dimensions=latent_dimensions, feature_size=12, variational=True
    )
    decoder_1 = architectures.Decoder(
        latent_dimensions=latent_dimensions, feature_size=10
    )
    decoder_2 = architectures.Decoder(
        latent_dimensions=latent_dimensions, feature_size=12
    )
    dvcca = DVCCA(
        latent_dimensions=latent_dimensions,
        encoders=[encoder_1, encoder_2],
        decoders=[decoder_1, decoder_2],
    )
    trainer = pl.Trainer(max_epochs=max_epochs, **trainer_kwargs)
    trainer.fit(dvcca, train_loader)



"""
Test whether linear simulated data has the correct correlations when we run cca on it
"""
from cca_zoo.data.simulated import LinearSimulatedData
import numpy as np


def test_data_correlation():
    data = LinearSimulatedData(
        view_features=[10, 10], latent_dims=5, correlation=[0.9, 0.8, 0.7, 0.6, 0.5]
    )
    x_train, y_train = data.sample(1000)
    x_test, y_test = data.sample(1000)
    from cca_zoo.classical import CCA

    model = CCA(latent_dimensions=5)
    model.fit((x_train, y_train))
    assert np.allclose(
        model.score((x_test, y_test)), np.array([0.9, 0.8, 0.7, 0.6, 0.5]), atol=0.1
    )



import torch

from cca_zoo.classical._iterative._ey import PLSEY
from cca_zoo.classical._iterative._base import BaseGradientLoop
from cca_zoo.classical._pls import PLSMixin


class PLSStochasticPower(PLSEY, PLSMixin):
    def _get_module(self, weights=None, k=None):
        return StochasticPowerLoopBase(
            weights=weights,
            k=k,
            learning_rate=self.learning_rate,
            optimizer_kwargs=self.optimizer_kwargs,
        )

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class StochasticPowerLoopBase(BaseGradientLoop):
    def __init__(self, weights=None, k=None, learning_rate=1e-3, optimizer_kwargs=None):
        super().__init__(
            weights=weights,
            k=k,
            learning_rate=learning_rate,
            optimizer_kwargs=optimizer_kwargs,
        )

    def training_step(self, batch, batch_idx):
        for weight in self.weights:
            weight.data = self._orth(weight)
        scores = self(batch["views"])
        # find the pairwise covariance between the scores
        cov = torch.cov(torch.hstack(scores).T)
        loss = torch.trace(cov[: scores[0].shape[1], scores[0].shape[1] :])
        self.log("train_loss", loss)
        return loss

    def on_train_end(self) -> None:
        for weight in self.weights:
            weight.data = self._orth(weight)

    @staticmethod
    def _orth(U):
        Qu, Ru = torch.linalg.qr(U)
        Su = torch.sign(torch.sign(torch.diag(Ru)) + 0.5)
        return Qu @ torch.diag(Su)



from typing import Iterable, Union

import numpy as np
import pytorch_lightning as pl

from cca_zoo.classical._iterative._base import BaseLoop
from cca_zoo.classical._iterative._deflation import BaseDeflation
from cca_zoo.classical._search import _softthreshold
from cca_zoo.utils import _process_parameter


class SCCA_Parkhomenko(BaseDeflation):
    r"""
    A class used to fit a sparse CCA (penalized CCA) model for two or more views.

    This model finds the linear projections of multiple views that maximize their pairwise correlations while enforcing sparsity constraints on the projection vectors.

    The objective function of sparse CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2 - \sum_i c_i\|w_i\|_1 \}\\

        \text{subject to:}

        w_i^Tw_i=1

    where :math:`c_i` are the sparsity parameters for each view.

    References
    ----------
    Parkhomenko, Elena, David Tritchler, and Joseph Beyene. "Sparse canonical correlation analysis with application to genomic data integration." Statistical applications in genetics and molecular biology 8.1 (2009).

    Examples
    --------
    >>> from cca_zoo.classical import SCCA_Parkhomenko
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_Parkhomenko(tau=[0.001,0.001],random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.81803527])
    """

    def __init__(
        self,
        latent_dimensions: int = 1,
        copy_data=True,
        random_state=None,
        deflation="cca",
        tau: Union[Iterable[float], float] = None,
        initialization: Union[str, callable] = "pls",
        tol: float = 1e-3,
        convergence_checking=False,
        patience=10,
        track=False,
        verbose=False,
    ):
        self.tau = tau
        super().__init__(
            latent_dimensions=latent_dimensions,
            copy_data=copy_data,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            convergence_checking=convergence_checking,
            patience=patience,
            track=track,
            verbose=verbose,
        )

    def _check_params(self):
        self.tau = _process_parameter("tau", self.tau, 0.0001, self.n_views_)
        if any(tau <= 0 for tau in self.tau):
            raise (
                "All regularisation parameters should be above 0. " f"tau=[{self.tau}]"
            )

    def _get_module(self, weights=None, k=None):
        return ParkhomenkoLoop(
            weights=weights,
            k=k,
            tau=self.tau,
            tol=self.tol,
        )

    def _more_tags(self):
        return {"multiview": True}


class ParkhomenkoLoop(BaseLoop):
    def __init__(self, weights, k=None, tau=None, tol=1e-3):
        super().__init__(weights=weights, k=k)
        self.tau = tau
        self.tol = tol

    def training_step(self, batch, batch_idx):
        scores = np.stack(self(batch["views"]))
        # Update each view using loop update function
        for view_index, view in enumerate(batch["views"]):
            # create a mask that is True for elements not equal to k along dim k
            mask = np.arange(scores.shape[0]) != view_index
            # apply the mask to scores and sum along dim k
            target = np.sum(scores[mask], axis=0)
            self.weights[view_index] = np.cov(
                np.hstack((batch["views"][view_index], target[:, np.newaxis])).T
            )[:-1, -1]
            self.weights[view_index] /= np.linalg.norm(self.weights[view_index])
            self.weights[view_index] = _softthreshold(
                self.weights[view_index], self.tau[view_index] / 2
            )
            self.weights[view_index] /= np.linalg.norm(self.weights[view_index])



import numpy as np
import pytest

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.classical import CCA


def test_PCCA():
    # some might not have access to jax/numpyro so leave this as an optional test locally.
    numpyro = pytest.importorskip("numpyro")
    from cca_zoo.probabilistic import ProbabilisticCCA

    np.random.seed(0)
    # Tests tensor CCA methods
    X, Y = LinearSimulatedData([5, 5]).sample(100)
    latent_dims = 1
    cca = CCA(latent_dimensions=latent_dims).fit([X, Y])
    pcca = ProbabilisticCCA(
        latent_dimensions=latent_dims, num_warmup=1000, num_samples=1000
    ).fit([X, Y])
    # Test that vanilla CCA and VCCA produce roughly similar latent space ie they are correlated
    assert (
        np.abs(
            np.corrcoef(
                cca.transform([X, Y])[1].T,
                pcca.posterior_samples["z"].mean(axis=0)[:, 0],
            )[0, 1]
        )
        > 0.9
    )



import numpy as np
import pytest
import scipy.sparse as sp
from sklearn.utils import check_random_state

from cca_zoo.classical import CCA, PLS, MCCA

n = 50
rng = check_random_state(0)
X = rng.rand(n, 10)
Y = rng.rand(n, 11)
Z = rng.rand(n, 12)
X_sp = sp.random(n, 10, density=0.5, random_state=rng)
Y_sp = sp.random(n, 11, density=0.5, random_state=rng)
# centre the data
X -= X.mean(axis=0)
Y -= Y.mean(axis=0)
Z -= Z.mean(axis=0)
X_sp -= X_sp.mean(axis=0)
Y_sp -= Y_sp.mean(axis=0)

latent_dims = 3
epochs = 250
batch_size = 10
learning_rate = 1e-1
random_state = 1


def scale_objective(obj):
    # log scale the objective if negative, otherwise nan
    obj = np.array(obj)
    return np.sign(obj) * np.log(np.abs(obj) + 1e-10)


def test_stochastic_pls():
    pytest.importorskip("torch")
    from torch import manual_seed

    from cca_zoo.classical import PLSEY, PLSSVD, PLSStochasticPower

    pls = PLS(latent_dimensions=3).fit((X, Y))
    manual_seed(42)
    plsey = PLSEY(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))
    plssvd = PLSSVD(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))
    spls = PLSStochasticPower(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))

    def scale_transform(model, X, Y):
        Zx, Zy = model.transform((X, Y))
        Zx /= np.linalg.norm(model.weights[0], axis=0, keepdims=True)
        Zy /= np.linalg.norm(model.weights[1], axis=0, keepdims=True)
        return np.abs(np.cov(Zx, Zy, rowvar=False)[:latent_dims, latent_dims:])

    pls_score = scale_transform(pls, X, Y)
    spls_score = scale_transform(spls, X, Y)
    plsey_score = scale_transform(plsey, X, Y)
    plssvd_score = scale_transform(plssvd, X, Y)
    # check all methods are similar to pls
    assert np.allclose(np.trace(pls_score), np.trace(spls_score), atol=1e-1)
    assert np.allclose(np.trace(pls_score), np.trace(plsey_score), atol=1e-1)
    assert np.allclose(np.trace(pls_score), np.trace(plssvd_score), atol=1e-1)


def test_stochastic_cca():
    pytest.importorskip("torch")
    from cca_zoo.classical import CCAEY, CCAGH, CCASVD

    cca = CCA(latent_dimensions=3).fit((X, Y))
    ccaey = CCAEY(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))
    ccagh = CCAGH(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))
    ccasvd = CCASVD(
        latent_dimensions=latent_dims,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
        random_state=random_state,
        track=True,
        verbose=False,
    ).fit((X, Y))
    cca_score = cca.score((X, Y))
    ccaey_score = ccaey.score((X, Y))
    ccagh_score = ccagh.score((X, Y))
    ccasvd_score = ccasvd.score((X, Y))
    # check all methods are similar to cca
    assert np.allclose(cca_score.sum(), ccaey_score.sum(), atol=1e-1)
    assert np.allclose(cca_score.sum(), ccagh_score.sum(), atol=1e-1)
    assert np.allclose(cca_score.sum(), ccasvd_score.sum(), atol=1e-1)
